#!/usr/bin/env python3

# THIS FILE IS PART OF THE CYLC SUITE ENGINE.
# Copyright (C) NIWA & British Crown (Met Office) & Contributors.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""cylc [info] get-suite-config [OPTIONS] ARGS

Print parsed suite configuration items, after runtime inheritance.

By default all settings are printed. For specific sections or items
use -i/--item and wrap sections in square brackets, e.g.:
   cylc get-suite-config --item '[scheduling]initial cycle point'
Multiple items can be retrieved at once.

By default, unset values are printed as an empty string, or (for
historical reasons) as "None" with -o/--one-line. These defaults
can be changed with the -n/--null-value option.

Example:
  |# SUITE.RC
  |[runtime]
  |    [[modelX]]
  |        [[[environment]]]
  |            FOO = foo
  |            BAR = bar

$ cylc get-suite-config --item=[runtime][modelX][environment]FOO SUITE
foo

$ cylc get-suite-config --item=[runtime][modelX][environment] SUITE
FOO = foo
BAR = bar

$ cylc get-suite-config --item=[runtime][modelX] SUITE
...
[[[environment]]]
    FOO = foo
    BAR = bar
..."""

import sys
from cylc.flow.remote import remrun
if remrun():
    sys.exit(0)

from cylc.flow.config import SuiteConfig
from cylc.flow.option_parsers import CylcOptionParser as COP
from cylc.flow.suite_files import parse_suite_arg
from cylc.flow.templatevars import load_template_vars
from cylc.flow.terminal import cli_function


def get_option_parser():
    parser = COP(__doc__, jset=True, prep=True, icp=True)

    parser.add_option(
        "-i", "--item", metavar="[SEC...]ITEM",
        help="Item or section to print (multiple use allowed).",
        action="append", dest="item", default=[])

    parser.add_option(
        "-r", "--sparse",
        help="Only print items explicitly set in the config files.",
        action="store_true", default=False, dest="sparse")

    parser.add_option(
        "-p", "--python",
        help="Print native Python format.",
        action="store_true", default=False, dest="pnative")

    parser.add_option(
        "-a", "--all-tasks",
        help="For [runtime] items (e.g. --item='script') report "
        "values for all tasks prefixed by task name.",
        action="store_true", default=False, dest="alltasks")

    parser.add_option(
        "-n", "--null-value",
        help="The string to print for unset values (default nothing).",
        metavar="STRING", action="store", default='', dest="none_str")

    parser.add_option(
        "-m", "--mark-up",
        help="Prefix each line with '!cylc!'.",
        action="store_true", default=False, dest="markup")

    parser.add_option(
        "-o", "--one-line",
        help="Print multiple single-value items at once.",
        action="store_true", default=False, dest="oneline")

    parser.add_option(
        "-t", "--tasks",
        help="Print the suite task list "
             "[DEPRECATED: use 'cylc list SUITE'].",
        action="store_true", default=False, dest="tasks")

    parser.add_option(
        "-u", "--run-mode",
        help="Get config for suite run mode.", action="store", default="live",
        dest="run_mode", choices=['live', 'dummy', 'simulation'])

    return parser


@cli_function(get_option_parser)
def main(parser, options, reg):
    suite, suiterc = parse_suite_arg(options, reg)

    if options.markup:
        prefix = '!cylc!'
    else:
        prefix = ''

    config = SuiteConfig(
        suite,
        suiterc,
        options,
        load_template_vars(options.templatevars, options.templatevars_file))
    if options.tasks:
        for task in config.get_task_name_list():
            print(prefix + task)
    elif options.alltasks:
        for task in config.get_task_name_list():
            items = ['[runtime][' + task + ']' + i for i in options.item]
            print(prefix + task, end=' ')
            config.pcfg.idump(
                items, options.sparse, options.pnative, prefix,
                options.oneline, none_str=options.none_str)
    else:
        config.pcfg.idump(
            options.item, options.sparse, options.pnative, prefix,
            options.oneline, none_str=options.none_str)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# THIS FILE IS PART OF THE CYLC SUITE ENGINE.
# Copyright (C) NIWA & British Crown (Met Office) & Contributors.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""The application control logic for Tui."""

import sys

import urwid
from urwid import html_fragment
from urwid.wimp import SelectableIcon

from cylc.flow.network.client import SuiteRuntimeClient
from cylc.flow.exceptions import (
    ClientError,
    ClientTimeout,
    SuiteStopped
)
from cylc.flow.task_state import (
    TASK_STATUSES_ORDERED,
    TASK_STATUS_RUNAHEAD,
    TASK_STATUS_SUBMITTED,
    TASK_STATUS_RUNNING,
    TASK_STATUS_FAILED,
)
from cylc.flow.tui.data import (
    QUERY
)
import cylc.flow.tui.overlay as overlay
from cylc.flow.tui import (
    BINDINGS,
    FORE,
    BACK,
    JOB_COLOURS,
    SUITE_COLOURS,
)
from cylc.flow.tui.tree import (
    find_closest_focus,
    translate_collapsing
)
from cylc.flow.tui.util import (
    compute_tree,
    dummy_flow,
    get_task_status_summary,
    get_workflow_status_str,
    render_node
)


urwid.set_encoding('utf8')  # required for unicode task icons

TREE_EXPAND_DEPTH = [2]


class TuiWidget(urwid.TreeWidget):
    """Display widget for tree nodes.

    Arguments:
        node (TuiNode):
            The root tree node.
        max_depth (int):
            Determines which nodes are unfolded by default.
            The maximum tree depth to unfold.

    """

    # allows leaf nodes to be selectable, otherwise the cursor
    # will skip rows when the user navigates
    unexpandable_icon = SelectableIcon(' ', 0)

    def __init__(self, node, max_depth=None):
        if not max_depth:
            max_depth = TREE_EXPAND_DEPTH[0]
        self._node = node
        self._innerwidget = None
        self.is_leaf = not node.get_child_keys()
        if max_depth > 0:
            self.expanded = node.get_depth() < max_depth
        else:
            self.expanded = True
        widget = self.get_indented_widget()
        urwid.WidgetWrap.__init__(self, widget)

    def selectable(self):
        """Return True if this node is selectable.

        Allow all nodes to be selectable apart from job information nodes.

        """
        return self.get_node().get_value()['type_'] != 'job_info'

    def _is_leaf(self):
        """Return True if this node has no children

        Note: the `is_leaf` attribute doesn't seem to give the right
              answer.

        """
        return (
            not hasattr(self, 'git_first_child')
            or not self.get_first_child()
        )

    def get_display_text(self):
        """Compute the text to display for a given node.

        Returns:
            (object) - Text content for the urwid.Text widget,
            may be a string, tuple or list, see urwid docs.

        """
        node = self.get_node()
        value = node.get_value()
        data = value['data']
        type_ = value['type_']
        return render_node(node, data, type_)

    def keypress(self, size, key):
        """Handle expand & collapse requests.

        Overridden from urwid.TreeWidget to change the behaviour
        of the left arrow key which urwid uses for navigation
        but which we think should be used for collapsing.

        """
        ret = self.__super.keypress(size, key)
        if ret in ('left',):
            self.expanded = False
            self.update_expanded_icon()
            # return None so that this keypress is not allowed to
            # propagate up the tree
            return
        return key

    def get_indented_widget(self):
        if self.is_leaf:

            self._innerwidget = urwid.Columns(
                [
                    ('fixed', 1, self.unexpandable_icon),
                    self.get_inner_widget()
                ],
                dividechars=1
            )
        return self.__super.get_indented_widget()


class TuiNode(urwid.TreeNode):
    """Data storage object for leaf nodes."""

    def load_widget(self):
        return TuiWidget(self)


class TuiParentNode(urwid.ParentNode):
    """Data storage object for interior/parent nodes."""

    def load_widget(self):
        return TuiWidget(self)

    def load_child_keys(self):
        # Note: keys are really indices.
        data = self.get_value()
        return range(len(data['children']))

    def load_child_node(self, key):
        """Return either an TuiNode or TuiParentNode"""
        childdata = self.get_value()['children'][key]
        if 'children' in childdata:
            childclass = TuiParentNode
        else:
            childclass = TuiNode
        return childclass(
            childdata,
            parent=self,
            key=key,
            depth=self.get_depth() + 1
        )


class TuiApp:
    """An application to display a single Cylc workflow.

    This is a single workflow view component (purposefully).

    Multi-suite functionality can be achieved via a GScan-esque
    tab/selection panel.

    Arguments:
        reg (str):
            Suite registration

    """

    UPDATE_INTERVAL = 1
    CLIENT_TIMEOUT = 1

    palette = [
        ('head', FORE, BACK),
        ('body', FORE, BACK),
        ('foot', 'white', 'dark blue'),
        ('key', 'light cyan', 'dark blue'),
        ('title', FORE, BACK, 'bold'),
        ('overlay', 'black', 'light gray'),
    ] + [  # job colours
        (f'job_{status}', colour, BACK)
        for status, colour in JOB_COLOURS.items()
    ] + [  # job colours for help screen
        (f'overlay_job_{status}', colour, 'light gray')
        for status, colour in JOB_COLOURS.items()
    ] + [  # suite state colours
        (f'suite_{status}',) + spec
        for status, spec in SUITE_COLOURS.items()
    ]

    def __init__(self, reg, screen=None):
        self.reg = reg
        self.client = None
        self.loop = None
        self.screen = None
        self.stack = 0
        self.tree_walker = None

        # create the template
        topnode = TuiParentNode(dummy_flow({'id': 'Loading...'}))
        self.listbox = urwid.TreeListBox(urwid.TreeWalker(topnode))
        header = urwid.Text('\n')
        footer = urwid.AttrWrap(
            # urwid.Text(self.FOOTER_TEXT),
            urwid.Text(list_bindings()),
            'foot'
        )
        self.view = urwid.Frame(
            urwid.AttrWrap(self.listbox, 'body'),
            header=urwid.AttrWrap(header, 'head'),
            footer=footer
        )
        self.filter_states = {
            state: True
            for state in TASK_STATUSES_ORDERED
            if state is not TASK_STATUS_RUNAHEAD
        }
        if isinstance(screen, html_fragment.HtmlGenerator):
            # the HtmlGenerator only captures one frame
            # so we need to pre-populate the GUI before
            # starting the event loop
            self.update()

    def main(self):
        """Start the event loop."""
        self.loop = urwid.MainLoop(
            self.view,
            self.palette,
            unhandled_input=self.unhandled_input,
            screen=self.screen
        )
        # schedule the first update
        self.loop.set_alarm_in(0, self._update)
        self.loop.run()

    def unhandled_input(self, key):
        """Catch key presses, uncaught events are passed down the chain."""
        if key in ('ctrl d',):
            raise urwid.ExitMainLoop()
        for binding in BINDINGS:
            # iterate through key bindings in order
            if key in binding['keys'] and binding['callback']:
                # if we get a match execute the callback
                # NOTE: if there is no callback then this binding is
                #       for documentation purposes only so we ignore it
                meth, *args = binding['callback']
                meth(self, *args)
                return

    def get_snapshot(self):
        """Contact the workflow, return a tree structure

        In the event of error contacting the suite the
        message is written to this Widget's header.

        Returns:
            dict if successful, else False

        """
        try:
            if not self.client:
                self.client = SuiteRuntimeClient(
                    self.reg,
                    timeout=self.CLIENT_TIMEOUT
                )
            data = self.client(
                'graphql',
                {
                    'request_string': QUERY,
                    'variables': {
                        # list of task states we want to see
                        'taskStates': [
                            state
                            for state, is_on in self.filter_states.items()
                            if is_on
                        ]
                    }
                }
            )
        except SuiteStopped:
            self.client = None
            return dummy_flow({
                'name': self.reg,
                'id': self.reg,
                'status': 'stopped',
                'stateTotals': {}
            })
        except (ClientError, ClientTimeout) as exc:
            # catch network / client errors
            self.set_header(('suite_error', str(exc)))
            return False

        if isinstance(data, list):
            # catch GraphQL errors
            try:
                message = data[0]['error']['message']
            except (IndexError, KeyError):
                message = str(data)
            self.set_header(('suite_error', message))
            return False

        if len(data['workflows']) != 1:
            # multiple workflows in returned data - shouldn't happen
            raise ValueError()

        return compute_tree(data['workflows'][0])

    @staticmethod
    def get_node_id(node):
        """Return a unique identifier for a node.

        Arguments:
            node (TuiNode): The node.

        Returns:
            str - Unique identifier

        """
        return node.get_value()['id_']

    def set_header(self, message):
        """Set the header message for this widget.

        Arguments:
            message (object):
                Text content for the urwid.Text widget,
                may be a string, tuple or list, see urwid docs.

        """
        # put in a one line gap
        if isinstance(message, list):
            message.append('\n')
        elif isinstance(message, tuple):
            message = (message[0], message[1] + '\n')
        else:
            message += '\n'
        self.view.header = urwid.Text(message)

    def _update(self, *_):
        try:
            self.update()
        except Exception as exc:
            sys.exit(exc)

    def update(self):
        """Refresh the data and redraw this widget.

        Preserves the current focus and collapse/expand state.

        """
        # update the data store
        # TODO: this can be done incrementally using deltas
        #       once this interface is available
        snapshot = self.get_snapshot()
        if snapshot is False:
            return False
        data = snapshot['data']

        # update the suite status message
        header = [get_workflow_status_str(data)]
        status_summary = get_task_status_summary(snapshot['data'])
        if status_summary:
            header.extend([' ('] + status_summary + [' )'])
        if not all(self.filter_states.values()):
            header.extend([' ', '*fitered* "R" to reset', ' '])
        self.set_header(header)

        # global update - the nuclear option - slow but simple
        # TODO: this can be done incrementally by adding and
        #       removing nodes from the existing tree
        topnode = TuiParentNode(snapshot)

        # NOTE: because we are nuking the tree we need to manually
        # preserve the focus and collapse status of tree nodes

        # record the old focus
        _, old_node = self.listbox._body.get_focus()

        # nuke the tree
        self.tree_walker = urwid.TreeWalker(topnode)
        self.listbox._set_body(self.tree_walker)

        # get the new focus
        _, new_node = self.listbox._body.get_focus()

        # preserve the focus or walk to the nearest parent
        closest_focus = find_closest_focus(self, old_node, new_node)
        self.listbox._body.set_focus(closest_focus)

        #  preserve the collapse/expand status of all nodes
        translate_collapsing(self, old_node, new_node)

        # schedule the next run of this update method
        if self.loop:
            self.loop.set_alarm_in(self.UPDATE_INTERVAL, self._update)

        return True

    def filter_by_task_state(self, filtered_state=None):
        """Filter tasks.

        Args:
            filtered_state (str):
                A task state to filter by or None.

        """
        self.filter_states = {
            state: (state == filtered_state) or not filtered_state
            for state in self.filter_states
        }
        return

    def open_overlay(self, fcn):
        self.create_overlay(*fcn(self))

    def create_overlay(self, widget, kwargs):
        """Open an overlay over the monitor.

        Args:
            widget (urwid.Widget):
                Widget to be placed inside the overlay.
            kwargs (dict):
                Dictionary of arguments to pass to the `urwid.Overlay`
                constructor.

                You will likely need to set `width` and `height` here.

                See `urwid` docs for details.

        """
        kwargs = {'width': 'pack', 'height': 'pack', **kwargs}
        overlay = urwid.Overlay(
            urwid.LineBox(
                urwid.AttrMap(
                    urwid.Frame(
                        urwid.Padding(
                            widget,
                            left=2,
                            right=2
                        ),
                        footer=urwid.Text('\n q to close')
                    ),
                    'overlay',
                )
            ),
            self.loop.widget,
            align='center',
            valign='middle',
            left=self.stack * 5,
            top=self.stack * 5,
            **kwargs,
        )
        self.loop.widget = overlay
        self.stack += 1

    def close_topmost(self):
        """Remove the topmost frame or uit the app if none present."""
        if self.stack > 0:
            self.loop.widget = self.loop.widget[0]
            self.stack -= 1
        else:
            raise urwid.ExitMainLoop()


BINDINGS.add_group(
    '',
    'Application Controls'
)
BINDINGS.bind(
    ('q',),
    '',
    'Quit',
    (TuiApp.close_topmost,)
)
BINDINGS.bind(
    ('h',),
    '',
    'Help',
    (TuiApp.open_overlay, overlay.help_info)
)
BINDINGS.bind(
    ('enter',),
    '',
    'Context',
    (TuiApp.open_overlay, overlay.context)
)

BINDINGS.add_group(
    'tree',
    'Expand/Collapse nodes',
)
BINDINGS.bind(
    ('-', '\u2190'),
    'tree',
    'Collapse',
    None  # this binding is for documentation only - handled by urwid
)
BINDINGS.bind(
    ('+', '\u2192'),
    'tree',
    'Expand',
    None  # this binding is for documentation only - handled by urwid
)

BINDINGS.add_group(
    'navigation',
    'Move within the tree'
)
BINDINGS.bind(
    ('\u2191',),
    'navigation',
    'Up',
    None  # this binding is for documentation only - handled by urwid
)
BINDINGS.bind(
    ('\u2193',),
    'navigation',
    'Down',
    None  # this binding is for documentation only - handled by urwid
)
BINDINGS.bind(
    ('\u21a5',),
    'navigation',
    'PageUp',
    None  # this binding is for documentation only - handled by urwid
)
BINDINGS.bind(
    ('\u21a7',),
    'navigation',
    'PageDown',
    None  # this binding is for documentation only - handled by urwid
)
BINDINGS.bind(
    ('Home',),
    'navigation',
    'Top',
    None  # this binding is for documentation only - handled by urwid
)
BINDINGS.bind(
    ('End',),
    'navigation',
    'Bottom',
    None  # this binding is for documentation only - handled by urwid
)

BINDINGS.add_group(
    'filter',
    'Filter by task state'
)
BINDINGS.bind(
    ('F',),
    'filter',
    'Select task states to filter by',
    (TuiApp.open_overlay, overlay.filter_task_state)
)
BINDINGS.bind(
    ('f',),
    'filter',
    'Show only failed tasks',
    (TuiApp.filter_by_task_state, TASK_STATUS_FAILED)
)
BINDINGS.bind(
    ('s',),
    'filter',
    'Show only submitted tasks',
    (TuiApp.filter_by_task_state, TASK_STATUS_SUBMITTED)
)
BINDINGS.bind(
    ('r',),
    'filter',
    'Show only running tasks',
    (TuiApp.filter_by_task_state, TASK_STATUS_RUNNING)
)
BINDINGS.bind(
    ('R',),
    'filter',
    'Reset task state filtering',
    (TuiApp.filter_by_task_state,)
)


def list_bindings():
    """Write out an in-line list of the key bindings."""
    ret = []
    for group, bindings in BINDINGS.list_groups():
        if group['name']:
            ret.append(f' {group["name"]}: ')
            for binding in bindings:
                for key in binding['keys']:
                    ret.append(('key', f'{key} '))
        else:
            # list each option in the default group individually
            for binding in bindings:
                ret.append(f'{binding["desc"].lower()}: ')
                ret.append(('key', binding["keys"][0]))
                ret.append(' ')
                ret.append(' ')
            ret.pop()  # remove surplus space
    return ret



# THIS FILE IS PART OF THE CYLC SUITE ENGINE.
# Copyright (C) NIWA & British Crown (Met Office) & Contributors.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""PBS batch system job submission and manipulation: multi-cluster variant.

Support PBS clients that front heterogeneous clusters where the Job ID returned
by qsub is <id>.<server>. PBS 13 qstat and qdel need <id>.<server>@<server>.
From PBS 14, the standard cylc PBS module works ("@<server>" is not needed).

So this PBS handler writes "job_id@server" to the job status file, and appends
"@server" to Job IDs returned by qstat, to matched the stored IDs.
"""

import re
from cylc.flow.batch_sys_handlers.pbs import PBSHandler


# Match and extract PBS Job ID of the form "<job>.<host>"
REC_JOB = re.compile(r'^ *(?P<job>[^ ]+?)\.(?P<host>[^ ]+) *$')
# Replace with "<job>.<host>@<host>"
REP_JOB = r'\g<job>.\g<host>@\g<host>'


class PBSMulticlusterHandler(PBSHandler):

    @classmethod
    def filter_poll_many_output(cls, out):
        """Extract and return Job IDs from qstat output.

        Ignore first two lines of qstat output (header lines).
        Manipulate Job IDs as described above, if matched, else leave as-is.
        """
        job_ids = []
        for line in out.splitlines()[2:]:
            job_ids.append(cls.manip_job_id(line))
        return job_ids

    @classmethod
    def manip_job_id(cls, job_id):
        """For job_id of the form "id.server", return job_id@server."""
        return re.sub(REC_JOB, REP_JOB, job_id.strip())


BATCH_SYS_HANDLER = PBSMulticlusterHandler()



#!/usr/bin/env python3

# THIS FILE IS PART OF THE CYLC SUITE ENGINE.
# Copyright (C) NIWA & British Crown (Met Office) & Contributors.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os


from cylc.flow.batch_sys_handlers.pbs import PBSHandler


class MyPBSHandler(PBSHandler):
    """For testing poll command connection refused."""
    @staticmethod
    def get_poll_many_cmd(_):
        """Always print PBSHandler.POLL_CANT_CONNECT_ERR to STDERR."""
        return os.path.join(os.path.dirname(__file__), 'badqstat')


BATCH_SYS_HANDLER = MyPBSHandler()



# THIS FILE IS PART OF THE CYLC SUITE ENGINE.
# Copyright (C) NIWA & British Crown (Met Office) & Contributors.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Module for unicode restrictions"""

import re


ENGLISH_REGEX_MAP = {
    r'\w': 'alphanumeric',
    r'\-': '-',
    r'\.': '.',
    r'\/': '/'
}


def regex_chars_to_text(chars):
    r"""Return a string representing a regex component.

    Examples:
        >>> regex_chars_to_text(['a', 'b', 'c'])
        ['a', 'b', 'c']
        >>> regex_chars_to_text([r'\-', r'\.', r'\/'])
        ['-', '.', '/']
        >>> regex_chars_to_text([r'\w'])
        ['alphanumeric']

    """
    return [
        ENGLISH_REGEX_MAP.get(char, char)
        for char in chars
    ]


def length(minimum, maximum):
    """Restrict character length.

    Example:
        >>> regex, message = length(0, 5)
        >>> message
        'must be between 0 and 5 characters long'
        >>> bool(regex.match('abcde'))
        True
        >>> bool(regex.match('abcdef'))
        False

    """
    return (
        re.compile(r'^.{%d,%d}$' % (minimum, maximum)),
        f'must be between {minimum} and {maximum} characters long'
    )


def allowed_characters(*chars):
    """Restrict permitted characters.

    Example:
        >>> regex, message = allowed_characters('a', 'b', 'c')
        >>> message
        'can only contain: a, b, c'
        >>> bool(regex.match('abc'))
        True
        >>> bool(regex.match('def'))
        False

    """
    return (
        re.compile(r'^[%s]+$' % ''.join(chars)),
        f'can only contain: {", ".join(regex_chars_to_text(chars))}'
    )


def not_starts_with(*chars):
    """Restrict first character.

    Example:
        >>> regex, message = not_starts_with('a', 'b', 'c')
        >>> message
        'can not start with: a, b, c'
        >>> bool(regex.match('def'))
        True
        >>> bool(regex.match('adef'))
        False

    """
    return (
        re.compile(r'^[^%s]' % ''.join(chars)),
        f'can not start with: {", ".join(regex_chars_to_text(chars))}'
    )


class UnicodeRuleChecker():

    RULES = []

    @classmethod
    def __init_subclass__(cls):
        cls.__doc__ = cls.__doc__ + '\n' if cls.__doc__ else ''
        cls.__doc__ += '\n' + '\n'.join([
            f'* {message}'
            for regex, message in cls.RULES
        ])

    @classmethod
    def validate(cls, string):
        """Run this collection of rules against the given string.

        Args:
            string (str):
                String to validate.

        Returns:
            tuple - (outcome, message)
            outcome (bool) - True if all patterns match.
            message (str) - User-friendly error message.

        """
        for rule, message in cls.RULES:
            if not rule.match(string):
                return (False, message)
        return (True, None)


class SuiteNameValidator(UnicodeRuleChecker):
    """The rules for valid suite names:"""

    RULES = [
        length(1, 254),
        not_starts_with(r'\.', r'\-'),
        allowed_characters(r'\w', r'\/', '_', '+', r'\-', r'\.', '@')
    ]



# THIS FILE IS PART OF THE CYLC SUITE ENGINE.
# Copyright (C) NIWA & British Crown (Met Office) & Contributors.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import pytest

from cylc.flow.exceptions import TriggerExpressionError
from cylc.flow.task_trigger import TaskTrigger, Dependency


def test_check_with_cycle_point():
    task_trigger = TaskTrigger('fake_task_name', '1', None, 'fakeOutput')
    actual = str(task_trigger)
    expected = 'fake_task_name[1]:fakeOutput'
    assert actual == expected


def test_check_with_no_cycle_point_with_offset():
    task_trigger = TaskTrigger('fake_task_name', None, 2, 'fakeOutput')
    actual = str(task_trigger)
    expected = 'fake_task_name[2]:fakeOutput'
    assert actual == expected


def test_check_with_no_cycle_point_or_offset():
    task_trigger = TaskTrigger('fake_task_name', None, None, 'fakeOutput')
    actual = str(task_trigger)
    expected = 'fake_task_name:fakeOutput'
    assert actual == expected


def test_check_for_false_suicide():
    task_trigger = TaskTrigger('fake_task_name', '1', None, 'fakeOutput')
    dependency = Dependency(
        [task_trigger, '&', task_trigger], [task_trigger], False)
    actual = str(dependency)
    expected = (
        '( fake_task_name[1]:fakeOutput ) ( & ) ( fake_task_name[1]'
        ':fakeOutput )')
    assert actual == expected


def test_check_for_true_suicide():
    task_trigger = TaskTrigger('fake_task_name', None, None, 'fakeOutput')
    dependency = Dependency(
        [task_trigger, '&', task_trigger], [task_trigger], True)
    actual = str(dependency)
    expected = (
        '! ( fake_task_name:fakeOutput ) ( & ) ( fake_task_name:fakeOutput )')
    assert actual == expected


def test_check_for_list_of_lists_exp():
    task_trigger = TaskTrigger('fake_task_name', None, None, 'fakeOutput')
    dependency = Dependency(
        [
            task_trigger,
            '&',
            ['task', '&', 'another_task']
        ],
        [task_trigger],
        False
    )
    actual = str(dependency)
    expected = (
        "( fake_task_name:fakeOutput ) ( & ) ['task', '&', 'another_task']")
    assert actual == expected


def test_check_exeption():
    with pytest.raises(TriggerExpressionError):
        TaskTrigger.get_trigger_name("Foo:Elephant")



#!/usr/bin/env python3
# THIS FILE IS PART OF THE CYLC SUITE ENGINE.
# Copyright (C) NIWA & British Crown (Met Office) & Contributors.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""cylc [task] tui REG

Open the terminal user interface (TUI) for the given suite.
"""
from textwrap import indent

from urwid import html_fragment

from cylc.flow.option_parsers import CylcOptionParser as COP
from cylc.flow.terminal import cli_function
from cylc.flow.tui import (
    TUI
)
from cylc.flow.tui.app import (
    TuiApp,
    TREE_EXPAND_DEPTH
    # ^ a nasty solution
)


__doc__ += indent(TUI, '           ')


def get_option_parser():
    parser = COP(
        __doc__,
        argdoc=[
            ('REG', 'Suite name')
        ],
        # auto_add=False,  NOTE: at present auto_add can not be turned off
        color=False
    )

    parser.add_option(
        '--display',
        help=(
            'Specify the display technology to use.'
            ' "raw" for interactive in-terminal display.'
            ' "html" for non-interactive html output.'
        ),
        action='store',
        choices=['raw', 'html'],
        default='raw',
    )
    parser.add_option(
        '--v-term-size',
        help=(
            'The virtual terminal size for non-interactive'
            '--display options.'
        ),
        action='store',
        default='80,24'
    )

    return parser


@cli_function(get_option_parser)
def main(_, options, reg):
    screen = None
    if options.display == 'html':
        TREE_EXPAND_DEPTH[0] = -1  # expand tree fully
        screen = html_fragment.HtmlGenerator()
        screen.set_terminal_properties(256)
        screen.register_palette(TuiApp.palette)
        html_fragment.screenshot_init(
            [tuple(map(int, options.v_term_size.split(',')))],
            []
        )

    try:
        TuiApp(reg, screen=screen).main()

        if options.display == 'html':
            for fragment in html_fragment.screenshot_collect():
                print(fragment)
    except KeyboardInterrupt:
        pass



#!/usr/bin/env python3

# THIS FILE IS PART OF THE CYLC SUITE ENGINE.
# Copyright (C) NIWA & British Crown (Met Office) & Contributors.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""cylc [prep] diff|compare [OPTIONS] SUITE1 SUITE2

Compare two suite definitions and display any differences.

Differencing is done after parsing the suite.rc files so it takes
account of default values that are not explicitly defined, it disregards
the order of configuration items, and it sees any include-file content
after inlining has occurred.

Files in the suite bin directory and other sub-directories of the
suite definition directory are not currently differenced."""

import sys
from cylc.flow.remote import remrun
if remrun():
    sys.exit(0)

from cylc.flow.option_parsers import CylcOptionParser as COP
from cylc.flow.config import SuiteConfig
from cylc.flow.suite_files import parse_suite_arg
from cylc.flow.templatevars import load_template_vars
from cylc.flow.terminal import cli_function

n_oone = 0
n_otwo = 0
n_diff = 0


def diffdict(one, two, oone, otwo, diff):
    global n_oone, n_otwo, n_diff
    # Recursively difference two dictionaries in which any element
    # may be another dictionary, keeping items that appear only
    # in one or the other, and items that appear in both but differ.
    for key in one:
        if key not in two:
            oone[key] = one[key]
            n_oone += 1
        elif one[key] != two[key]:
            if isinstance(one[key], dict):
                for item in oone, otwo, diff:
                    if key not in item:
                        item[key] = {}
                diffdict(one[key], two[key], oone[key], otwo[key], diff[key])
            else:
                diff[key] = (one[key], two[key])
                n_diff += 1

    for key in two:
        if key not in one:
            otwo[key] = two[key]
            n_otwo += 1


def prdict(dct, arrow='<', section='', level=0, diff=False, nested=False):
    """Recursively print, in pseudo 'diff' format, the contents of
    one of the three dictionaries populated by the diffdict() function
    above (any element may itself be a dictionary).
    """

    if section != '':
        prfx = section + ' '
    else:
        prfx = ''

    if section == '':
        sctn = '(top)'
    else:
        sctn = section

    foo = False

    for key in dct:
        if isinstance(dct[key], dict):
            lvl = level + 1
            if nested:
                pre = prfx + '\n' + '   ' * lvl
            else:
                pre = prfx
            prdict(dct[key], arrow,
                   pre + '[' * lvl + str(key) + ']' * lvl, lvl,
                   diff, nested)
        else:
            if not foo:
                if nested:
                    print('  ', sctn)
                else:
                    print('\n  ', sctn)
                foo = True

            if diff:
                print(' <  ', key, '=', dct[key][0])
                print(' >  ', key, '=', dct[key][1])
            else:
                print(' ' + arrow + '  ', key, '=', dct[key])


def get_option_parser():
    parser = COP(
        __doc__, jset=True, prep=True, icp=True,
        argdoc=[('SUITE1', 'Suite name or path'),
                ('SUITE2', 'Suite name or path')])

    parser.add_option(
        "-n", "--nested",
        help="print suite.rc section headings in nested form.",
        action="store_true", default=False, dest="nested")

    return parser


@cli_function(get_option_parser)
def main(parser, options, *args):
    suite1, suite1rc = parse_suite_arg(options, args[0])
    suite2, suite2rc = parse_suite_arg(options, args[1])
    if suite1 == suite2:
        parser.error("You can't diff a single suite.")
    print("Parsing %s (%s)" % (suite1, suite1rc))
    template_vars = load_template_vars(
        options.templatevars, options.templatevars_file)
    config1 = SuiteConfig(suite1, suite1rc, options, template_vars).cfg
    print("Parsing %s (%s)" % (suite2, suite2rc))
    config2 = SuiteConfig(
        suite2, suite2rc, options, template_vars, is_reload=True).cfg

    if config1 == config2:
        print("Suite definitions %s and %s are identical" % (suite1, suite2))
        sys.exit(0)

    print("Suite definitions %s and %s differ" % (suite1, suite2))

    suite1_only = {}
    suite2_only = {}
    diff_1_2 = {}

    diffdict(config1, config2, suite1_only, suite2_only, diff_1_2)

    if n_oone > 0:
        print()
        msg = str(n_oone) + ' items only in ' + suite1 + ' (<)'
        print(msg)
        prdict(suite1_only, '<', nested=options.nested)

    if n_otwo > 0:
        print()
        msg = str(n_otwo) + ' items only in ' + suite2 + ' (>)'
        print(msg)
        prdict(suite2_only, '>', nested=options.nested)

    if n_diff > 0:
        print()
        msg = (str(n_diff) + ' common items differ ' +
               suite1 + '(<) ' + suite2 + '(>)')
        print(msg)
        prdict(diff_1_2, '', diff=True, nested=options.nested)


if __name__ == "__main__":
    main()



# THIS FILE IS PART OF THE CYLC SUITE ENGINE.
# Copyright (C) NIWA & British Crown (Met Office) & Contributors.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""Provide a function to report modification to broadcast settings."""


BAD_OPTIONS_FMT = "\n  --%s=%s"
BAD_OPTIONS_TITLE = "No broadcast to cancel/clear for these options:"
BAD_OPTIONS_TITLE_SET = ("Rejected broadcast: settings are not"
                         " compatible with the suite")
CHANGE_FMT = "\n%(change)s [%(namespace)s.%(point)s] %(key)s=%(value)s"
CHANGE_PREFIX_CANCEL = "-"
CHANGE_PREFIX_SET = "+"
CHANGE_TITLE_CANCEL = "Broadcast cancelled:"
CHANGE_TITLE_SET = "Broadcast set:"


def get_broadcast_bad_options_report(bad_options, is_set=False):
    """Return a string to report bad options for broadcast cancel/clear."""
    if not bad_options:
        return None
    if is_set:
        msg = BAD_OPTIONS_TITLE_SET
    else:
        msg = BAD_OPTIONS_TITLE
    for key, values in sorted(bad_options.items()):
        for value in values:
            if isinstance(value, tuple) or isinstance(value, list):
                value_str = ""
                values = list(value)
                while values:
                    val = values.pop(0)
                    if values:
                        value_str += "[" + val + "]"
                    else:
                        value_str += val
            else:
                value_str = value
            msg += BAD_OPTIONS_FMT % (key, value_str)
    return msg


def get_broadcast_change_iter(modified_settings, is_cancel=False):
    """Return an iterator of broadcast changes.

    Each broadcast change is a dict with keys:
    change, point, namespace, key, value

    """
    if not modified_settings:
        return
    if is_cancel:
        change = CHANGE_PREFIX_CANCEL
    else:
        change = CHANGE_PREFIX_SET
    for modified_setting in sorted(modified_settings,
                                   key=lambda x: (x[0], x[1])):
        # sorted by (point, namespace)
        point, namespace, setting = modified_setting
        value = setting
        keys_str = ""
        while isinstance(value, dict):
            key, value = list(value.items())[0]
            if isinstance(value, dict):
                keys_str += "[" + key + "]"
            else:
                keys_str += key
                yield {
                    "change": change,
                    "point": point,
                    "namespace": namespace,
                    "key": keys_str,
                    "value": str(value)}


def get_broadcast_change_report(modified_settings, is_cancel=False):
    """Return a string for reporting modification to broadcast settings."""
    if not modified_settings:
        return ""
    if is_cancel:
        msg = CHANGE_TITLE_CANCEL
    else:
        msg = CHANGE_TITLE_SET
    for broadcast_change in get_broadcast_change_iter(
            modified_settings, is_cancel):
        msg += CHANGE_FMT % broadcast_change
    return msg



# THIS FILE IS PART OF THE CYLC SUITE ENGINE.
# Copyright (C) 2008-2019 NIWA & British Crown (Met Office) & Contributors.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Creates authentication keys for use in testing"""

from cylc.flow.suite_files import (
    create_server_keys,
    get_suite_srv_dir,
    KeyInfo,
    KeyOwner,
    KeyType,
    remove_keys_on_server)
from cylc.flow.task_remote_cmd import (
    remove_keys_on_platform, create_platform_keys)


def setup_keys(suite_name):
    suite_srv_dir = get_suite_srv_dir(suite_name)
    server_keys = {
        "client_public_key": KeyInfo(
            KeyType.PUBLIC,
            KeyOwner.CLIENT,
            suite_srv_dir=suite_srv_dir),
        "client_private_key": KeyInfo(
            KeyType.PRIVATE,
            KeyOwner.CLIENT,
            suite_srv_dir=suite_srv_dir),
        "server_public_key": KeyInfo(
            KeyType.PUBLIC,
            KeyOwner.SERVER,
            suite_srv_dir=suite_srv_dir),
        "server_private_key": KeyInfo(
            KeyType.PRIVATE,
            KeyOwner.SERVER,
            suite_srv_dir=suite_srv_dir)
    }
    remove_keys_on_server(server_keys)
    remove_keys_on_platform(suite_srv_dir)
    create_server_keys(server_keys, suite_srv_dir)
    create_platform_keys(suite_srv_dir)



# THIS FILE IS PART OF THE CYLC SUITE ENGINE.
# Copyright (C) NIWA & British Crown (Met Office) & Contributors.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""Cylc site and user configuration file spec."""

import os
import re

from cylc.flow import LOG
from cylc.flow import __version__ as CYLC_VERSION
from cylc.flow.hostuserutil import get_user_home, is_remote_user
from cylc.flow.network.authorisation import Priv
from cylc.flow.parsec.config import ParsecConfig, ConfigNode as Conf
from cylc.flow.parsec.exceptions import ParsecError
from cylc.flow.parsec.upgrade import upgrader
from cylc.flow.parsec.validate import (
    DurationFloat, CylcConfigValidator as VDR, cylc_config_validate)

# Nested dict of spec items.
# Spec value is [value_type, default, allowed_2, allowed_3, ...]
# where:
# - value_type: value type (compulsory).
# - default: the default value (optional).
# - allowed_2, ...: the only other allowed values of this setting (optional).
with Conf('flow.rc', desc='''
    The global configuration which defines default Cylc Flow settings
    for a user or site.

    To view your global config run::

       $ cylc get-global-config --sparse


    Cylc will attempt to load the global configuration (flow.rc) from two
    locations:

    * ``/etc/cylc/flow/<CYLC_VERSION>/flow.rc``
    * ``~/.cylc/flow/<CYLC_VERSION>/flow.rc``

    If both files are present files will be loaded in this order so those
    lower down the list may override settings from those higher up.

    To override the default configuration path use ``CYLC_CONF_PATH``.

    .. note::

       The ``flow.rc`` file can be templated using Jinja2 variables.
       See :ref:`Jinja`.
''') as SPEC:

    # suite
    Conf('process pool size', VDR.V_INTEGER, 4, desc='''
        Maximum number of concurrent processes used to execute external job
        submission, event handlers, and job poll and kill commands - see
        :ref:`Managing External Command Execution`.
    ''')
    Conf('process pool timeout', VDR.V_INTERVAL, DurationFloat(600), desc='''
        Interval after which long-running commands in the process pool will be
        killed - see :ref:`Managing External Command Execution`.

        .. note::
           The default is set quite high to avoid killing important
           processes when the system is under load.
    ''')
    # client
    Conf('disable interactive command prompts', VDR.V_BOOLEAN, True, desc='''
        Commands that intervene in running suites can be made to ask for
        confirmation before acting. Some find this annoying and ineffective as
        a safety measure, however, so command prompts are disabled by default.
    ''')
    # suite
    Conf('run directory rolling archive length', VDR.V_INTEGER, -1, desc='''
        The number of old run directory trees to retain at start-up.
    ''')

    # suite
    with Conf('cylc', desc='''
        Default values for entries in the suite.rc ``[cylc]`` section.
    '''):
        Conf('UTC mode', VDR.V_BOOLEAN, desc='''
                Default for :cylc:conf:`suite.rc[cylc]UTC mode`.
        ''')
        Conf('task event mail interval', VDR.V_INTERVAL, DurationFloat(300),
             desc='''
                Default for
                :cylc:conf:`suite.rc[cylc]task event mail interval`.
        ''')

        with Conf('events', desc='''
            You can define site defaults for each of the following options,
            details of which can be found under
            :cylc:conf:`suite.rc[cylc][events]`.
        '''):
            Conf('handlers', VDR.V_STRING_LIST)
            Conf('handler events', VDR.V_STRING_LIST)
            Conf('mail events', VDR.V_STRING_LIST)
            Conf('mail from', VDR.V_STRING)
            Conf('mail smtp', VDR.V_STRING)
            Conf('mail to', VDR.V_STRING)
            Conf('mail footer', VDR.V_STRING)
            Conf('startup handler', VDR.V_STRING_LIST)
            Conf('timeout handler', VDR.V_STRING_LIST)
            Conf('inactivity handler', VDR.V_STRING_LIST)
            Conf('shutdown handler', VDR.V_STRING_LIST)
            Conf('aborted handler', VDR.V_STRING_LIST)
            Conf('stalled handler', VDR.V_STRING_LIST)
            Conf('timeout', VDR.V_INTERVAL)
            Conf('inactivity', VDR.V_INTERVAL)
            Conf('abort on timeout', VDR.V_BOOLEAN)
            Conf('abort on inactivity', VDR.V_BOOLEAN)
            Conf('abort on stalled', VDR.V_BOOLEAN)

        with Conf('main loop', desc='''
            Configuration of the Cylc Scheduler's main loop.
        '''):
            Conf('plugins', VDR.V_STRING_LIST, ['health check'], desc='''
                Configure the default main loop plugins to use when
                starting up new suites.
            ''')

            with Conf('<plugin name>', desc='''
                Configure a main loop plugin.
            ''') as MainLoopPlugin:
                Conf('interval', VDR.V_INTERVAL, desc='''
                    The interval with which this plugin is run.
                ''')
                Conf('foo', VDR.V_STRING, default='X')

            with Conf('health check', meta=MainLoopPlugin, desc='''
                Checks the integrity of the suite run directory.
            '''):
                Conf('interval', VDR.V_INTERVAL, DurationFloat(600), desc='''
                    The interval with which this plugin is run.
                ''')

    # suite
    with Conf('suite logging', desc='''
        The suite event log, held under the suite run directory, is maintained
        as a rolling archive. Logs are rolled over (backed up and started anew)
        when they reach a configurable limit size.
    '''):
        Conf('rolling archive length', VDR.V_INTEGER, 5, desc='''
            How many rolled logs to retain in the archive.
        ''')
        Conf('maximum size in bytes', VDR.V_INTEGER, 1000000, desc='''
            Suite event logs are rolled over when they reach this file size.
        ''')

    # general
    with Conf('documentation', desc='''
        Documentation locations for the ``cylc doc`` command.
    '''):
        Conf('local', VDR.V_STRING, '', desc='''
            Path where the cylc documentation will appear if built locally.
        ''')
        Conf('online', VDR.V_STRING,
             'http://cylc.github.io/doc/built-sphinx/index.html', desc='''
            URL of the online cylc documentation.
        ''')
        Conf('cylc homepage', VDR.V_STRING, 'http://cylc.github.io/', desc='''
            URL of the cylc internet homepage, with links to documentation for
            the latest official release.
        ''')

    # general
    with Conf('document viewers', desc='''
        PDF and HTML viewers can be launched by cylc to view the
        documentation.
    '''):
        Conf('html', VDR.V_STRING, 'firefox', desc='''
            Your preferred web browser.
        ''')

    # client
    with Conf('editors', desc='''
        Choose your favourite text editor for editing suite configurations.
    '''):
        Conf('terminal', VDR.V_STRING, 'vim', desc='''
            The editor to be invoked by the cylc command line interface.

            Examples::

                emacs -nw
                vi -f
        ''')
        Conf('gui', VDR.V_STRING, 'gvim -f', desc='''
            GUI Text editor to be invoked by Cylc:

            Examples::

                gvim -g
                emacs
        ''')

    # job platforms
    with Conf('job platforms'):
        with Conf('<platform name>'):
            Conf('batch system', VDR.V_STRING, 'background')
            Conf('batch submit command template', VDR.V_STRING)
            Conf('shell', VDR.V_STRING, '/bin/bash')
            Conf('run directory', VDR.V_STRING, '$HOME/cylc-run', desc='''
                The number of old run directory trees to retain at start-up.
            ''')
            Conf('work directory', VDR.V_STRING, '$HOME/cylc-run', desc='''
                The top level for suite work and share directories. Can contain
                ``$HOME`` or ``$USER`` but not other environment variables (the
                item cannot actually be evaluated by the shell on HOST before
                use, but the remote home directory is where ``rsync`` and
                ``ssh`` naturally land, and the remote username is known by the
                suite server program).

                Example::

                   /nfs/data/$USER/cylc-run
            ''')
            Conf('suite definition directory', VDR.V_STRING)
            Conf('task communication method',
                 VDR.V_STRING, 'zmq', options=['zmq', 'poll'], desc='''
                The means by which task progress messages are reported back to
                the running suite.

                Options:

                zmq
                   Direct client-server TCP communication via network ports
                poll
                   The suite polls for the status of tasks (no task messaging)
            ''')
            # TODO ensure that it is possible to over-ride the following three
            # settings in suite config.
            Conf('submission polling intervals', VDR.V_INTERVAL_LIST, desc='''
                Cylc can also poll submitted jobs to catch problems that
                prevent the submitted job from executing at all, such as
                deletion from an external batch scheduler queue. Routine
                polling is done only for the polling ``task communication
                method`` unless suite-specific polling is configured in
                the suite configuration. A list of interval values can be
                specified as for execution polling but a single value
                is probably sufficient for job submission polling.

                Example::

                   5*PT1M, 10*PT5M
            ''')
            Conf('submission retry delays', VDR.V_INTERVAL_LIST, None)
            Conf('execution polling intervals', VDR.V_INTERVAL_LIST, desc='''
                Cylc can poll running jobs to catch problems that prevent task
                messages from being sent back to the suite, such as hard job
                kills, network outages, or unplanned task host shutdown.
                Routine polling is done only for the polling *task
                communication method* (below) unless suite-specific polling is
                configured in the suite configuration.  A list of interval
                values can be specified, with the last value used repeatedly
                until the task is finished - this allows more frequent polling
                near the beginning and end of the anticipated task run time.
                Multipliers can be used as shorthand as in the example below.

                Example::

                   5*PT1M, 10*PT5M
            ''')
            Conf('execution time limit polling intervals',
                 VDR.V_INTERVAL_LIST, desc='''
                The intervals between polling after a task job (submitted to
                the relevant batch system on the relevant host) exceeds its
                execution time limit. The default setting is PT1M, PT2M, PT7M.
                The accumulated times (in minutes) for these intervals will be
                roughly 1, 1 + 2 = 3 and 1 + 2 + 7 = 10 after a task job
                exceeds its execution time limit.
            ''')
            Conf('scp command',
                 VDR.V_STRING, 'scp -oBatchMode=yes -oConnectTimeout=10',
                 desc='''
                A string for the command used to copy files to a remote host.
                This is not used on the suite host unless you run local tasks
                under another user account. The value is assumed to be ``scp``
                with some initial options or a command that implements a
                similar interface to ``scp``.
            ''')
            Conf('ssh command',
                 VDR.V_STRING, 'ssh -oBatchMode=yes -oConnectTimeout=10',
                 desc='''
                A string for the command used to invoke commands on this host.
                This is not used on the suite host unless you run local tasks
                under another user account.  The value is assumed to be ``ssh``
                with some initial options or a command that implements a
                similar interface to ``ssh``.
            ''')
            Conf('use login shell', VDR.V_BOOLEAN, True, desc='''
                Whether to use a login shell or not for remote command
                invocation. By default cylc runs remote ssh commands using a
                login shell:

                .. code-block:: bash

                   ssh user@host 'bash --login cylc ...'

                which will source ``/etc/profile`` and ``~/.profile`` to set up
                the user environment.  However, for security reasons some
                institutions do not allow unattended commands to start login
                shells, so you can turn off this behaviour to get:

                .. code-block:: bash

                   ssh user@host 'cylc ...'

                which will use the default shell on the remote machine,
                sourcing ``~/.bashrc`` (or ``~/.cshrc``) to set up the
                environment.
            ''')
            Conf('remote hosts', VDR.V_STRING_LIST)
            Conf('cylc executable', VDR.V_STRING, 'cylc', desc='''
                The ``cylc`` executable on a remote host.

                .. note::

                   This should normally point to the cylc multi-version wrapper
                   on the host, not ``bin/cylc`` for a specific installed
                   version.

                Specify a full path if ``cylc`` is not in ``$PATH`` when it is
                invoked via ``ssh`` on this host.
            ''')
            Conf('global init-script', VDR.V_STRING, desc='''
                If specified, the value of this setting will be inserted to
                just before the ``init-script`` section of all job scripts that
                are to be submitted to the specified remote host.
            ''')
            Conf('copyable environment variables', VDR.V_STRING_LIST, '',
                 desc='''
                A list containing the names of the environment variables that
                can and/or need to be copied from the suite server program to a
                job.
            ''')
            Conf('retrieve job logs', VDR.V_BOOLEAN, desc='''
                Global default for
                :cylc:conf:`suite.rc[runtime][<namespace>][remote]retrieve job
                logs`.
            ''')
            Conf('retrieve job logs command', VDR.V_STRING, 'rsync -a',
                 desc='''
                If ``rsync -a`` is unavailable or insufficient to retrieve job
                logs from a remote host, you can use this setting to specify a
                suitable command.
            ''')
            Conf('retrieve job logs max size', VDR.V_STRING, desc='''
                Global default for the
                :cylc:conf:`suite.rc[runtime][<namespace>][remote]retrieve job
                logs max size`.
                the specified host.
            ''')
            Conf('retrieve job logs retry delays', VDR.V_INTERVAL_LIST,
                 desc='''
                Global default for the
                :cylc:conf:`suite.rc[runtime][<namespace>][remote]retrieve job
                logs retry delays`.
                setting for the specified host.
            ''')
            Conf('task event handler retry delays', VDR.V_INTERVAL_LIST,
                 desc='''
                Host specific default for
                :cylc:conf:`suite.rc[runtime][<namespace>][events]handler retry
                delays`.
            ''')
            Conf('tail command template',
                 VDR.V_STRING, 'tail -n +1 -F %(filename)s', desc='''
                A command template (with ``%(filename)s`` substitution) to
                tail-follow job logs on HOST, by ``cylc cat-log``. You are
                unlikely to need to override this.
            ''')
            Conf('err tailer', VDR.V_STRING, desc='''
                A command template (with ``%(job_id)s`` substitution) that can
                be used to tail-follow the stderr stream of a running job if
                SYSTEM does not use the normal log file location while the job
                is running.  This setting overrides
                :cylc:conf:`[..]tail command template`.

                Examples::

                   # for PBS
                   qcat -f -e %(job_id)s
            ''')
            Conf('out tailer', VDR.V_STRING, desc='''
                A command template (with ``%(job_id)s`` substitution) that can
                be used to tail-follow the stdout stream of a running job if
                SYSTEM does not use the normal log file location while the job
                is running.  This setting overrides
                :cylc:conf:`[..]tail command template`.

                Examples::

                   # for PBS
                   qcat -f -o %(job_id)s
            ''')
            Conf('err viewer', VDR.V_STRING, desc='''
                A command template (with ``%(job_id)s`` substitution) that can
                be used to view the stderr stream of a running job if SYSTEM
                does not use the normal log file location while the job is
                running.

                Examples::

                   # for PBS
                   qcat -e %(job_id)s
            ''')
            Conf('out viewer', VDR.V_STRING, desc='''
                A command template (with ``%(job_id)s`` substitution) that can
                be used to view the stdout stream of a running job if SYSTEM
                does not use the normal log file location while the job is
                running.

                Examples::

                   # for PBS
                   qcat -o %(job_id)s
            ''')
            Conf('job name length maximum', VDR.V_INTEGER, desc='''
                The maximum length for job name acceptable by a batch system on
                a given host.  Currently, this setting is only meaningful for
                PBS jobs. For example, PBS 12 or older will fail a job submit
                if the job name has more than 15 characters; whereas PBS 13
                accepts up to 236 characters.
            ''')
            Conf('owner', VDR.V_STRING)

    # Platform Groups
    with Conf('platform groups'):
        with Conf('<group>'):
            Conf('platforms', VDR.V_STRING_LIST)

    # task
    with Conf('hosts'):
        with Conf('localhost'):
            Conf('run directory', VDR.V_STRING, '$HOME/cylc-run')
            Conf('work directory', VDR.V_STRING, '$HOME/cylc-run')
            Conf('task communication method',
                 VDR.V_STRING, 'default', options=['default', 'ssh', 'poll'])
            Conf('submission polling intervals', VDR.V_INTERVAL_LIST)
            Conf('execution polling intervals', VDR.V_INTERVAL_LIST)
            Conf('scp command',
                 VDR.V_STRING, 'scp -oBatchMode=yes -oConnectTimeout=10')
            Conf('ssh command',
                 VDR.V_STRING, 'ssh -oBatchMode=yes -oConnectTimeout=10')
            Conf('use login shell', VDR.V_BOOLEAN, True)
            Conf('cylc executable', VDR.V_STRING, 'cylc')
            Conf('global init-script', VDR.V_STRING)
            Conf('copyable environment variables', VDR.V_STRING_LIST)
            Conf('retrieve job logs', VDR.V_BOOLEAN)
            Conf('retrieve job logs command', VDR.V_STRING, 'rsync -a')
            Conf('retrieve job logs max size', VDR.V_STRING)
            Conf('retrieve job logs retry delays', VDR.V_INTERVAL_LIST)
            Conf('task event handler retry delays', VDR.V_INTERVAL_LIST)
            Conf('tail command template',
                 VDR.V_STRING, 'tail -n +1 -F %(filename)s')
            with Conf('batch systems'):
                with Conf('<batch system name>'):
                    Conf('err tailer', VDR.V_STRING)
                    Conf('out tailer', VDR.V_STRING)
                    Conf('err viewer', VDR.V_STRING)
                    Conf('out viewer', VDR.V_STRING)
                    Conf('job name length maximum', VDR.V_INTEGER)
                    Conf('execution time limit polling intervals',
                         VDR.V_INTERVAL_LIST)

        with Conf('<hostname glob>'):
            Conf('run directory', VDR.V_STRING)
            Conf('work directory', VDR.V_STRING)
            Conf('task communication method',
                 VDR.V_STRING, 'default', options=['default', 'ssh', 'poll'])
            Conf('submission polling intervals', VDR.V_INTERVAL_LIST)
            Conf('execution polling intervals', VDR.V_INTERVAL_LIST)
            Conf('scp command', VDR.V_STRING)
            Conf('ssh command', VDR.V_STRING)
            Conf('use login shell', VDR.V_BOOLEAN)
            Conf('cylc executable', VDR.V_STRING)
            Conf('global init-script', VDR.V_STRING)
            Conf('copyable environment variables', VDR.V_STRING_LIST)
            Conf('retrieve job logs', VDR.V_BOOLEAN)
            Conf('retrieve job logs command', VDR.V_STRING)
            Conf('retrieve job logs max size', VDR.V_STRING)
            Conf('retrieve job logs retry delays', VDR.V_INTERVAL_LIST)
            Conf('task event handler retry delays', VDR.V_INTERVAL_LIST)
            Conf('tail command template', VDR.V_STRING)
            with Conf('batch systems'):
                with Conf('<batch system name>'):
                    Conf('err tailer', VDR.V_STRING)
                    Conf('out tailer', VDR.V_STRING)
                    Conf('out viewer', VDR.V_STRING)
                    Conf('err viewer', VDR.V_STRING)
                    Conf('job name length maximum', VDR.V_INTEGER)
                    Conf('execution time limit polling intervals',
                         VDR.V_INTERVAL_LIST)

    # task
    with Conf('task events', desc='''
        Global site/user defaults for
        :cylc:conf:`suite.rc[runtime][<namespace>][events]`.
    '''):
        Conf('execution timeout', VDR.V_INTERVAL)
        Conf('handlers', VDR.V_STRING_LIST)
        Conf('handler events', VDR.V_STRING_LIST)
        Conf('handler retry delays', VDR.V_INTERVAL_LIST, None)
        Conf('mail events', VDR.V_STRING_LIST)
        Conf('mail from', VDR.V_STRING)
        Conf('mail retry delays', VDR.V_INTERVAL_LIST)
        Conf('mail smtp', VDR.V_STRING)
        Conf('mail to', VDR.V_STRING)
        Conf('submission timeout', VDR.V_INTERVAL)

    # client
    with Conf('test battery', desc='''
        Settings for the automated development tests.

        .. note::
           The test battery reads ``flow-tests.rc`` instead of the normal
           site/user global config files (from the same locations, however).
    '''):
        Conf('remote host with shared fs', VDR.V_STRING, desc='''
            The name of a remote host that sees the same HOME file system as
            the host running the test battery.
        ''')
        Conf('remote host', VDR.V_STRING, desc='''
            Host name of a remote account that does not see the same home
            directory as the account running the test battery - see also
            "remote owner" below).
        ''')
        Conf('remote owner', VDR.V_STRING, desc='''
            User name of a remote account that does not see the same home
            directory as the account running the test battery - see also
            "remote host" above).
        ''')

        with Conf('batch systems', desc='''
            Settings for testing supported batch systems (job submission
            methods). The tests for a batch system are only performed if the
            batch system is available on the test host or a remote host
            accessible via SSH from the test host.
        '''):

            with Conf('<batch system name>', desc='''
                SYSTEM is the name of a supported batch system with automated
                tests.  This can currently be "loadleveler", "lsf", "pbs",
                "sge" and/or "slurm".
            '''):
                Conf('host', VDR.V_STRING, desc='''
                    The name of a host where commands for this batch system is
                    available. Use "localhost" if the batch system is available
                    on the host running the test battery. Any specified remote
                    host should be accessible via SSH from the host running the
                    test battery.
                ''')
                Conf('out viewer', VDR.V_STRING, desc='''
                    The command template (with ``%(job_id)s`` substitution)
                    for testing the run time stdout viewer functionality for
                    this batch system.
                ''')
                Conf('err viewer', VDR.V_STRING, desc='''
                    The command template (with ``%(job_id)s`` substitution)
                    for testing the run time stderr viewer functionality for
                    this batch system.
                ''')

                with Conf('directives', desc='''
                    The minimum set of directives that must be supplied to the
                    batch system on the site to initiate jobs for the tests.
                '''):
                    Conf('<directive>', VDR.V_STRING)

    # suite
    with Conf('suite host self-identification', desc='''
        The suite host's identity must be determined locally by cylc and passed
        to running tasks (via ``$CYLC_SUITE_HOST``) so that task messages can
        target the right suite on the right host.

        .. todo
           Is it conceivable that different remote task hosts at the same site
           might see the suite host differently? If so we would need to be able
           to override the target in suite configurations.
    '''):
        Conf('method', VDR.V_STRING, 'name',
             options=['name', 'address', 'hardwired'], desc='''
            This item determines how cylc finds the identity of the suite host.
            For the default *name* method cylc asks the suite host for its host
            name. This should resolve on remote task hosts to the IP address of
            the suite host; if it doesn't, adjust network settings or use one
            of the other methods. For the *address* method, cylc attempts to
            use a special external "target address" to determine the IP address
            of the suite host as seen by remote task hosts.  And finally, as a
            last resort, you can choose the *hardwired* method and manually
            specify the host name or IP address of the suite host.

            Options:

            name
               Self-identified host name.
            address
               Automatically determined IP address (requires *target*).
            hardwired
               Manually specified host name or IP address (requires *host*).
        ''')
        Conf('target', VDR.V_STRING, 'google.com', desc='''
            This item is required for the *address* self-identification method.
            If your suite host sees the internet, a common address such as
            ``google.com`` will do; otherwise choose a host visible on your
            intranet.
        ''')
        Conf('host', VDR.V_STRING, desc='''
            Use this item to explicitly set the name or IP address of the suite
            host if you have to use the *hardwired* self-identification method.
        ''')

    # suite
    with Conf('authentication', desc='''
        Authentication of client programs with suite server programs can be
        configured here, and overridden in suites if necessary with
        :cylc:conf:`suite.rc[cylc][authentication]`.

        The suite-specific passphrase must be installed on a user's account to
        authorize full control privileges (see
        :ref:`ConnectionAuthentication`). In the future we plan to move to a
        more traditional user account model so that each authorized user can
        have their own password.
    '''):
        # Allow owners to grant public shutdown rights at the most, not full
        # control.
        Conf(
            'public',
            VDR.V_STRING,
            default=Priv.STATE_TOTALS.name.lower().replace('_', '-'),
            options=[
                level.name.lower().replace('_', '-')
                for level in [
                    Priv.IDENTITY, Priv.DESCRIPTION,
                    Priv.STATE_TOTALS, Priv.READ, Priv.SHUTDOWN
                ]
            ],
            desc='''
                This sets the client privilege level for public access - i.e.
                no suite passphrase required.
            '''
        )

    # suite
    with Conf('suite servers', desc='''
        Configure allowed suite hosts and ports for starting up (running or
        restarting) suites. Additionally configure host selection settings
        specifying how to determine the most suitable run host at any given
        time from those configured.
    '''):
        Conf('run hosts', VDR.V_SPACELESS_STRING_LIST, desc='''
            A list of allowed suite run hosts. One of these hosts will be
            appointed for a suite to start up on if an explicit host is not
            provided as an option to a ``run`` or ``restart`` command.
        ''')
        Conf('run ports', VDR.V_INTEGER_LIST, list(range(43001, 43101)),
             desc='''
            A list of allowed ports for Cylc to use to run suites.
        ''')
        Conf('condemned hosts', VDR.V_ABSOLUTE_HOST_LIST, desc='''
            Hosts specified in ``condemned hosts`` will not be considered as
            suite run hosts. If suites are already running on ``condemned
            hosts`` they will be automatically shutdown and restarted (see
            :ref:`auto-stop-restart`).
        ''')
        Conf('auto restart delay', VDR.V_INTERVAL, desc='''
            Relates to Cylc's auto stop-restart mechanism (see
            :ref:`auto-stop-restart`).  When a host is set to automatically
            shutdown/restart it will first wait a random period of time between
            zero and ``auto restart delay`` seconds before beginning the
            process. This is to prevent large numbers of suites from restarting
            simultaneously.
        ''')
        Conf('ranking', VDR.V_STRING)


def upg(cfg, descr):
    """Upgrader."""
    u = upgrader(cfg, descr)

    u.obsolete('6.4.1', ['test battery', 'directives'])
    u.obsolete('6.11.0', ['state dump rolling archive length'])
    # Roll over is always done.
    u.obsolete('7.8.0', ['suite logging', 'roll over at start-up'])
    u.obsolete('7.8.1', ['documentation', 'local index'])
    u.obsolete('7.8.1', ['documentation', 'files', 'pdf user guide'])
    u.obsolete('7.8.1', ['documentation', 'files',
                         'single-page html user guide'])
    u.deprecate('7.8.1',
                ['documentation', 'files', 'multi-page html user guide'],
                ['documentation', 'local'])
    u.deprecate('8.0.0',
                ['documentation', 'files', 'html index'],
                ['documentation', 'local'])
    u.deprecate('8.0.0',
                ['documentation', 'urls', 'internet homepage'],
                ['documentation', 'cylc homepage'])
    u.obsolete('8.0.0', ['suite servers', 'scan hosts'])
    u.obsolete('8.0.0', ['suite servers', 'scan ports'])
    u.obsolete('8.0.0', ['communication'])
    u.obsolete('8.0.0', ['temporary directory'])
    u.obsolete('8.0.0', ['task host select command timeout'])
    u.obsolete('8.0.0', ['xtrigger function timeout'])
    u.obsolete('8.0.0', ['enable run directory housekeeping'])
    u.obsolete('8.0.0', ['task messaging'])

    u.upgrade()


class GlobalConfig(ParsecConfig):
    """
    Handle global (all suites) site and user configuration for cylc.
    User file values override site file values.
    """

    _DEFAULT = None
    _HOME = os.getenv('HOME') or get_user_home()
    CONF_BASENAME = "flow.rc"
    SITE_CONF_DIR = os.path.join(os.sep, 'etc', 'cylc', 'flow', CYLC_VERSION)
    USER_CONF_DIR = os.path.join(_HOME, '.cylc', 'flow', CYLC_VERSION)

    @classmethod
    def get_inst(cls, cached=True):
        """Return a GlobalConfig instance.

        Args:
            cached (bool):
                If cached create if necessary and return the singleton
                instance, else return a new instance.
        """
        if not cached:
            # Return an up-to-date global config without affecting the
            # singleton.
            new_instance = cls(SPEC, upg, validator=cylc_config_validate)
            new_instance.load()
            return new_instance
        elif not cls._DEFAULT:
            cls._DEFAULT = cls(SPEC, upg, validator=cylc_config_validate)
            cls._DEFAULT.load()
        return cls._DEFAULT

    def load(self):
        """Load or reload configuration from files."""
        self.sparse.clear()
        self.dense.clear()
        LOG.debug("Loading site/user config files")
        conf_path_str = os.getenv("CYLC_CONF_PATH")
        if conf_path_str:
            # Explicit config file override.
            fname = os.path.join(conf_path_str, self.CONF_BASENAME)
            if os.access(fname, os.F_OK | os.R_OK):
                self.loadcfg(fname, upgrader.USER_CONFIG)
        elif conf_path_str is None:
            # Use default locations.
            for conf_dir, conf_type in [
                    (self.SITE_CONF_DIR, upgrader.SITE_CONFIG),
                    (self.USER_CONF_DIR, upgrader.USER_CONFIG)]:
                fname = os.path.join(conf_dir, self.CONF_BASENAME)
                if not os.access(fname, os.F_OK | os.R_OK):
                    continue
                try:
                    self.loadcfg(fname, conf_type)
                except ParsecError as exc:
                    if conf_type == upgrader.SITE_CONFIG:
                        # Warn on bad site file (users can't fix it).
                        LOG.warning(
                            'ignoring bad %s %s:\n%s', conf_type, fname, exc)
                    else:
                        # Abort on bad user file (users can fix it).
                        LOG.error('bad %s %s', conf_type, fname)
                        raise
        # (OK if no flow.rc is found, just use system defaults).
        self._transform()

    def get_host_item(self, item, host=None, owner=None, replace_home=False,
                      owner_home=None):
        """This allows hosts with no matching entry in the config file
        to default to appropriately modified localhost settings."""

        cfg = self.get()

        # (this may be called with explicit None values for localhost
        # and owner, so we can't use proper defaults in the arg list)
        if not host:
            # if no host is given the caller is asking about localhost
            host = 'localhost'

        # is there a matching host section?
        host_key = None
        if host in cfg['hosts']:
            # there's an entry for this host
            host_key = host
        else:
            # try for a pattern match
            for cfg_host in cfg['hosts']:
                if re.match(cfg_host, host):
                    host_key = cfg_host
                    break
        modify_dirs = False
        if host_key is not None:
            # entry exists, any unset items under it have already
            # defaulted to modified localhost values (see site cfgspec)
            value = cfg['hosts'][host_key][item]
        else:
            # no entry so default to localhost and modify appropriately
            value = cfg['hosts']['localhost'][item]
            modify_dirs = True
        if value is not None and 'directory' in item:
            if replace_home or modify_dirs:
                # Replace local home dir with $HOME for eval'n on other host.
                value = value.replace(self._HOME, '$HOME')
            elif is_remote_user(owner):
                # Replace with ~owner for direct access via local filesys
                # (works for standard cylc-run directory location).
                if owner_home is None:
                    owner_home = os.path.expanduser('~%s' % owner)
                value = value.replace(self._HOME, owner_home)
        if item == "task communication method" and value == "default":
            # Translate "default" to client-server comms: "zmq"
            value = 'zmq'
        return value

    def _transform(self):
        """Transform various settings.

        Host item values of None default to modified localhost values.
        Expand environment variables and ~ notations.

        Ensure os.environ['HOME'] is defined with the correct value.
        """
        cfg = self.get()

        for host in cfg['hosts']:
            if host == 'localhost':
                continue
            for item, value in cfg['hosts'][host].items():
                if value is None:
                    newvalue = cfg['hosts']['localhost'][item]
                else:
                    newvalue = value
                if newvalue and 'directory' in item:
                    # replace local home dir with $HOME for evaluation on other
                    # host
                    newvalue = newvalue.replace(self._HOME, '$HOME')
                cfg['hosts'][host][item] = newvalue

        # Expand environment variables and ~user in LOCAL file paths.
        if 'HOME' not in os.environ:
            os.environ['HOME'] = self._HOME
        cfg['documentation']['local'] = os.path.expandvars(
            cfg['documentation']['local'])
        for key, val in cfg['hosts']['localhost'].items():
            if val and 'directory' in key:
                cfg['hosts']['localhost'][key] = os.path.expandvars(val)



# THIS FILE IS PART OF THE CYLC SUITE ENGINE.
# Copyright (C) NIWA & British Crown (Met Office) & Contributors.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from tempfile import NamedTemporaryFile, SpooledTemporaryFile, TemporaryFile,\
    TemporaryDirectory
import unittest

from pathlib import Path

from cylc.flow.subprocctx import SubProcContext
from cylc.flow.subprocpool import SubProcPool, _XTRIG_FUNCS, get_func


class TestSubProcPool(unittest.TestCase):

    def test_get_temporary_file(self):
        """Test SubProcPool.get_temporary_file."""
        self.assertIsInstance(
            SubProcPool.get_temporary_file(), SpooledTemporaryFile)

    def test_run_command_returns_0(self):
        """Test basic usage, command returns 0"""
        ctx = SubProcContext('truth', ['true'])
        SubProcPool.run_command(ctx)
        self.assertEqual(ctx.err, '')
        self.assertEqual(ctx.out, '')
        self.assertEqual(ctx.ret_code, 0)

    def test_run_command_returns_1(self):
        """Test basic usage, command returns 1"""
        ctx = SubProcContext('lies', ['false'])
        SubProcPool.run_command(ctx)
        self.assertEqual(ctx.err, '')
        self.assertEqual(ctx.out, '')
        self.assertEqual(ctx.ret_code, 1)

    def test_run_command_writes_to_out(self):
        """Test basic usage, command writes to STDOUT"""
        ctx = SubProcContext('parrot', ['echo', 'pirate', 'urrrr'])
        SubProcPool.run_command(ctx)
        self.assertEqual(ctx.err, '')
        self.assertEqual(ctx.out, 'pirate urrrr\n')
        self.assertEqual(ctx.ret_code, 0)

    def test_run_command_writes_to_err(self):
        """Test basic usage, command writes to STDERR"""
        ctx = SubProcContext(
            'parrot2', ['bash', '-c', 'echo pirate errrr >&2'])
        SubProcPool.run_command(ctx)
        self.assertEqual(ctx.err, 'pirate errrr\n')
        self.assertEqual(ctx.out, '')
        self.assertEqual(ctx.ret_code, 0)

    def test_run_command_with_stdin_from_str(self):
        """Test STDIN from string"""
        ctx = SubProcContext('meow', ['cat'], stdin_str='catches mice.\n')
        SubProcPool.run_command(ctx)
        self.assertEqual(ctx.err, '')
        self.assertEqual(ctx.out, 'catches mice.\n')
        self.assertEqual(ctx.ret_code, 0)

    def test_run_command_with_stdin_from_unicode(self):
        """Test STDIN from string with Unicode"""
        ctx = SubProcContext('meow', ['cat'], stdin_str='喵\n')
        SubProcPool.run_command(ctx)
        self.assertEqual(ctx.err, '')
        self.assertEqual(ctx.out, '喵\n')
        self.assertEqual(ctx.ret_code, 0)

    def test_run_command_with_stdin_from_handle(self):
        """Test STDIN from a single opened file handle"""
        handle = TemporaryFile()
        handle.write('catches mice.\n'.encode('UTF-8'))
        handle.seek(0)
        ctx = SubProcContext('meow', ['cat'], stdin_files=[handle])
        SubProcPool.run_command(ctx)
        self.assertEqual(ctx.err, '')
        self.assertEqual(ctx.out, 'catches mice.\n')
        self.assertEqual(ctx.ret_code, 0)
        handle.close()

    def test_run_command_with_stdin_from_path(self):
        """Test STDIN from a single file path"""
        handle = NamedTemporaryFile()
        handle.write('catches mice.\n'.encode('UTF-8'))
        handle.seek(0)
        ctx = SubProcContext('meow', ['cat'], stdin_files=[handle.name])
        SubProcPool.run_command(ctx)
        self.assertEqual(ctx.err, '')
        self.assertEqual(ctx.out, 'catches mice.\n')
        self.assertEqual(ctx.ret_code, 0)
        handle.close()

    def test_run_command_with_stdin_from_handles(self):
        """Test STDIN from multiple file handles"""
        handles = []
        for txt in ['catches mice.\n', 'eat fish.\n']:
            handle = TemporaryFile()
            handle.write(txt.encode('UTF-8'))
            handle.seek(0)
            handles.append(handle)
        ctx = SubProcContext('meow', ['cat'], stdin_files=handles)
        SubProcPool.run_command(ctx)
        self.assertEqual(ctx.err, '')
        self.assertEqual(ctx.out, 'catches mice.\neat fish.\n')
        self.assertEqual(ctx.ret_code, 0)
        for handle in handles:
            handle.close()

    def test_run_command_with_stdin_from_paths(self):
        """Test STDIN from multiple file paths"""
        handles = []
        for txt in ['catches mice.\n', 'eat fish.\n']:
            handle = NamedTemporaryFile()
            handle.write(txt.encode('UTF-8'))
            handle.seek(0)
            handles.append(handle)
        ctx = SubProcContext(
            'meow', ['cat'], stdin_files=[handle.name for handle in handles])
        SubProcPool.run_command(ctx)
        self.assertEqual(ctx.err, '')
        self.assertEqual(ctx.out, 'catches mice.\neat fish.\n')
        self.assertEqual(ctx.ret_code, 0)
        for handle in handles:
            handle.close()

    def test_xfunction(self):
        """Test xtrigger function import."""
        with TemporaryDirectory() as temp_dir:
            python_dir = Path(temp_dir, "lib", "python")
            python_dir.mkdir(parents=True)
            the_answer_file = python_dir / "the_answer.py"
            with the_answer_file.open(mode="w") as f:
                f.write("""the_answer = lambda: 42""")
                f.flush()
            fn = get_func("the_answer", temp_dir)
            result = fn()
            self.assertEqual(42, result)

    def test_xfunction_cache(self):
        """Test xtrigger function import cache."""
        with TemporaryDirectory() as temp_dir:
            python_dir = Path(temp_dir, "lib", "python")
            python_dir.mkdir(parents=True)
            amandita_file = python_dir / "amandita.py"
            with amandita_file.open(mode="w") as f:
                f.write("""amandita = lambda: 'chocolate'""")
                f.flush()
            fn = get_func("amandita", temp_dir)
            result = fn()
            self.assertEqual('chocolate', result)

            # is in the cache
            self.assertTrue('amandita' in _XTRIG_FUNCS)
            # returned from cache
            self.assertEqual(fn, get_func("amandita", temp_dir))
            del _XTRIG_FUNCS['amandita']
            # is not in the cache
            self.assertFalse('amandita' in _XTRIG_FUNCS)

    def test_xfunction_import_error(self):
        """Test for error on importing a xtrigger function.

        To prevent the test eventually failing if the test function is added
        and successfully imported, we use an invalid module name as per Python
        spec.
        """
        with TemporaryDirectory() as temp_dir:
            with self.assertRaises(ModuleNotFoundError):
                get_func("invalid-module-name", temp_dir)

    def test_xfunction_attribute_error(self):
        """Test for error on looking for an attribute in a xtrigger script."""
        with TemporaryDirectory() as temp_dir:
            python_dir = Path(temp_dir, "lib", "python")
            python_dir.mkdir(parents=True)
            the_answer_file = python_dir / "the_sword.py"
            with the_answer_file.open(mode="w") as f:
                f.write("""the_droid = lambda: 'excalibur'""")
                f.flush()
            with self.assertRaises(AttributeError):
                get_func("the_sword", temp_dir)


if __name__ == '__main__':
    unittest.main()



# THIS FILE IS PART OF THE CYLC SUITE ENGINE.
# Copyright (C) NIWA & British Crown (Met Office) & Contributors.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import tempfile
import unittest

import jinja2

from cylc.flow.parsec.jinja2support import *


class TestJinja2support(unittest.TestCase):

    def test_raise_helper(self):
        message = 'Ops'
        error_type = 'CRITICAL'
        with self.assertRaises(Exception) as cm:
            raise_helper(message=message)

        ex = cm.exception
        self.assertEqual("Jinja2 Error: Ops", str(ex))

        with self.assertRaises(Exception) as cm:
            raise_helper(message=message, error_type=error_type)

        ex = cm.exception
        self.assertEqual("Jinja2 CRITICAL: Ops", str(ex))

    def test_assert_helper(self):
        assert_helper(logical=True, message="Doesn't matter")  # harmless

        with self.assertRaises(Exception):
            assert_helper(logical=False, message="Doesn't matter")

    def test_jinja2environment(self):
        # create a temp directory, in the temp directory, to prevent
        # issues running multiple test suites in parallel
        temp_directory = tempfile.mkdtemp(prefix='cylc', suffix='test_jinja2')
        filters_dir = os.path.join(temp_directory, 'Jinja2Filters')
        os.mkdir(filters_dir)
        with open(os.path.join(filters_dir, "min.py"), "w") as tf:
            tf.write("def min():\n    raise ArithmeticError('UP!')")
            tf.seek(0)
            dir_ = temp_directory
            env = jinja2environment(dir_)
            # our jinja env contains the following keys in the global namespace
            self.assertTrue('environ' in env.globals)
            self.assertTrue('raise' in env.globals)
            self.assertTrue('assert' in env.globals)

            with self.assertRaises(ArithmeticError) as cm:
                # jinja2environment must have loaded the function from the .py
                env.filters['min']()
            self.assertEqual('UP!', str(cm.exception))

    def test_jinja2process(self):
        lines = ["skipped", "My name is {{ name }}", ""]
        variables = {'name': 'Cylc'}
        template_dir = tempfile.gettempdir()

        r = jinja2process(lines, template_dir, variables)

        self.assertEqual(['My name is Cylc'], r)

    def test_jinja2process_missing_variables(self):
        lines = ["skipped", "My name is {{ name }}", ""]
        template_dir = tempfile.gettempdir()

        with self.assertRaises(Jinja2Error) as exc:
            jinja2process(lines, template_dir, template_vars=None)
            self.assertIn('jinja2.UndefinedError', str(exc))

    def test_pymoduleloader(self):
        temp_directory = tempfile.mkdtemp(prefix='cylc', suffix='test_jinja2')
        filters_dir = os.path.join(temp_directory, 'Jinja2filters')
        os.mkdir(filters_dir)
        with tempfile.NamedTemporaryFile(dir=filters_dir, suffix=".py") as tf:
            tf.write(
                "def jinja2jinja()\n    raise Exception('It works!')".encode())
            tf.seek(0)
            dir_ = temp_directory
            env = jinja2environment(dir_)

            module_loader = PyModuleLoader()
            template = module_loader.load(environment=env, name='sys')
            self.assertEqual(sys.path, template.module.path)

            template2 = module_loader.load(
                environment=env, name='__python__.sys')

            self.assertEqual(template.module.path, template2.module.path)

    def test_pymoduleloader_invalid_module(self):
        temp_directory = tempfile.mkdtemp(prefix='cylc', suffix='test_jinja2')
        filters_dir = os.path.join(temp_directory, 'Jinja2filters')
        os.mkdir(filters_dir)
        with tempfile.NamedTemporaryFile(dir=filters_dir, suffix=".py") as tf:
            tf.write(
                "def jinja2jinja()\n    raise Exception('It works!')".encode())
            tf.seek(0)
            dir_ = temp_directory
            env = jinja2environment(dir_)

            module_loader = PyModuleLoader()
            with self.assertRaises(jinja2.TemplateNotFound):
                module_loader.load(environment=env, name='no way jose')


if __name__ == '__main__':
    unittest.main()



# THIS FILE IS PART OF THE CYLC SUITE ENGINE.
# Copyright (C) NIWA & British Crown (Met Office) & Contributors.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from cylc.flow.cycling.loader import get_point_relative
from cylc.flow.exceptions import TriggerExpressionError
from cylc.flow.prerequisite import Prerequisite
from cylc.flow.task_outputs import (
    TASK_OUTPUT_EXPIRED, TASK_OUTPUT_SUBMITTED, TASK_OUTPUT_SUBMIT_FAILED,
    TASK_OUTPUT_STARTED, TASK_OUTPUT_SUCCEEDED, TASK_OUTPUT_FAILED)

# Task trigger names (e.g. foo:fail => bar).
# Can use "foo:fail => bar" or "foo:failed => bar", etc.
_ALT_TRIGGER_NAMES = {
    TASK_OUTPUT_EXPIRED: ["expire"],
    TASK_OUTPUT_SUBMITTED: ["submit"],
    TASK_OUTPUT_SUBMIT_FAILED: ["submit-fail"],
    TASK_OUTPUT_STARTED: ["start"],
    TASK_OUTPUT_SUCCEEDED: ["succeed"],
    TASK_OUTPUT_FAILED: ["fail"],
}


class TaskTrigger(object):
    """Class representing an upstream dependency.

    Args:
        task_name (str): The name of the upstream task.
        abs_cycle_point (cylc.flow.cycling.PointBase): The cycle point of the
            upstream dependency if it is an absolute dependency.
            e.g. foo[^] => ...). Else `None`.
        cycle_point_offset (str): String representing the offset of the
            upstream task (e.g. -P1D) if this dependency is not an absolute
            one. Else None.
        output (str): The task state / output for this trigger e.g. succeeded.

    """

    __slots__ = ['task_name', 'abs_cycle_point', 'cycle_point_offset',
                 'output']

    def __init__(self, task_name, abs_cycle_point, cycle_point_offset,
                 output):
        self.task_name = task_name
        self.abs_cycle_point = abs_cycle_point
        self.cycle_point_offset = cycle_point_offset
        self.output = output

    def get_point(self, point):
        """Return the point of the output to which this TaskTrigger pertains.

        Args:
            point (cylc.flow.cycling.PointBase): The cycle point of the
                dependent task.

        Returns:
            cylc.flow.cycling.PointBase: The cycle point of the dependency.

        """
        if self.abs_cycle_point:
            point = self.abs_cycle_point
        elif self.cycle_point_offset:
            point = get_point_relative(
                self.cycle_point_offset, point)
        return point

    def __str__(self):
        if self.abs_cycle_point:
            return '%s[%s]:%s' % (self.task_name, self.abs_cycle_point,
                                  self.output)
        elif self.cycle_point_offset:
            return '%s[%s]:%s' % (self.task_name, self.cycle_point_offset,
                                  self.output)
        else:
            return '%s:%s' % (self.task_name, self.output)

    @staticmethod
    def get_trigger_name(trigger_name):
        """Standardise a trigger name.

        E.g. 'foo:fail' becomes 'foo:failed' etc.

        """
        for standard_name, alt_names in _ALT_TRIGGER_NAMES.items():
            if trigger_name == standard_name or trigger_name in alt_names:
                return standard_name
        raise TriggerExpressionError(
            f"Illegal task trigger name: {trigger_name}")


class Dependency(object):
    """A graph dependency in its abstract form.

    Used to generate cylc.flow.prerequisite.Prerequisite objects.

    Args:
        exp (list): A (nested) list of TaskTrigger objects and conditional
            characters representing the dependency. E.G: "foo & bar" would be
            [<TaskTrigger("foo")>, "&", <TaskTrigger("Bar")>].
        task_triggers (set): A set of TaskTrigger objects contained in the
            expression (exp).
        suicide (bool): True if this is a suicide trigger else False.

    """

    __slots__ = ['_exp', 'task_triggers', 'suicide']

    def __init__(self, exp, task_triggers, suicide):
        self._exp = exp
        self.task_triggers = tuple(task_triggers)  # More memory efficient.
        self.suicide = suicide

    def get_prerequisite(self, point, tdef):
        """Generate a Prerequisite object from this dependency.

        Args:
            point (cylc.flow.cycling.PointBase): The cycle point at which to
                generate the Prerequisite for.
            tdef (cylc.flow.taskdef.TaskDef): The TaskDef of the dependent
                task.

        Returns:
            cylc.flow.prerequisite.Prerequisite

        """
        # Create Prerequisite.
        cpre = Prerequisite(point, tdef.start_point)

        # Loop over TaskTrigger instances.
        for task_trigger in self.task_triggers:
            if task_trigger.cycle_point_offset is not None:
                # Inter-cycle trigger - compute the trigger's cycle point from
                # its offset.
                prereq_offset_point = get_point_relative(
                    task_trigger.cycle_point_offset, point)
                if prereq_offset_point > point:
                    # Update tdef.max_future_prereq_offset.
                    prereq_offset = prereq_offset_point - point
                    if (tdef.max_future_prereq_offset is None or
                            (prereq_offset >
                             tdef.max_future_prereq_offset)):
                        tdef.max_future_prereq_offset = (
                            prereq_offset)
                pre_initial = ((prereq_offset_point < tdef.start_point) &
                               (point >= tdef.start_point))
                cpre.add(task_trigger.task_name,
                         task_trigger.get_point(point),
                         task_trigger.output,
                         pre_initial)
            else:
                # Trigger is within the same cycle point.
                # Register task message with Prerequisite object.
                cpre.add(task_trigger.task_name,
                         task_trigger.get_point(point),
                         task_trigger.output)
        cpre.set_condition(self.get_expression(point))
        return cpre

    def get_expression(self, point):
        """Return the expression as a string.

        Args:
            point (cylc.flow.cycling.PointBase): The cycle point at which to
                generate the expression string for.

        Return:
            string: The expression as a parsable string in the cylc graph
            format.

        """
        return ''.join(self._stringify_list(self._exp, point))

    def __str__(self):
        ret = []
        if self.suicide:
            ret.append('!')
        for item in self._exp:
            if isinstance(item, list):
                ret.append(str(item))
            else:
                ret.append('( %s )' % str(item))
        return ' '.join(ret)

    @classmethod
    def _stringify_list(cls, nested_expr, point):
        """Stringify a nested list of TaskTrigger objects."""
        ret = []
        for item in nested_expr:
            if isinstance(item, TaskTrigger):
                ret.append(Prerequisite.MESSAGE_TEMPLATE % (
                    item.task_name, item.get_point(point), item.output))
            elif isinstance(item, list):
                ret.extend(['('] + cls._stringify_list(item, point) + [')'])
            else:
                ret.append(item)
        return ret



# THIS FILE IS PART OF THE CYLC SUITE ENGINE.
# Copyright (C) NIWA & British Crown (Met Office) & Contributors.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""Client for suite runtime API."""

from functools import partial
import os
from shutil import which
import socket
import sys
from typing import Union

import zmq
import zmq.asyncio

from cylc.flow import LOG
from cylc.flow.exceptions import (
    ClientError,
    ClientTimeout,
    SuiteServiceFileError,
    SuiteStopped
)
from cylc.flow.network import (
    encode_,
    decode_,
    get_location,
    ZMQSocketBase
)
from cylc.flow.network.server import PB_METHOD_MAP
from cylc.flow.suite_files import detect_old_contact_file


class SuiteRuntimeClient(ZMQSocketBase):
    """Initiate a client to the workflow server API.

    Initiates the REQ part of a ZMQ REQ-REP pair.

    This class contains the logic for the ZMQ message interface and client -
    server communication.

    Determine host and port from the contact file unless provided.

    If there is no socket bound to the specified host/port the client will
    bail after ``timeout`` seconds.

    Args:
        suite (str):
            Name of the suite to connect to.
        owner (str):
            Owner of suite, defaults to $USER.
        host (str):
            Overt need to check contact file if provided along with the
            port.
        port (int):
            The port on the aforementioned host to connect to.
        timeout (float):
            Set the default timeout in seconds. The default is
            ``ZMQClient.DEFAULT_TIMEOUT``.
            Note the default timeout can be overridden for individual requests.
        timeout_handler (function):
            Optional function which runs before ClientTimeout is raised.
            This provides an interface for raising more specific exceptions in
            the event of a communication timeout.
        header (dict):
            Request "header" data to attach to each request.

    Usage:
        Call endpoints using ``ZMQClient.__call__``.

    Message interface:
        * Accepts responses of the format: {"data": {...}}
        * Accepts error in the format: {"error": {"message": MSG}}
        * Returns requests of the format: {"command": CMD,
          "args": {...}}

    Raises:
        ClientError: if the suite is not running.

    Call server "endpoints" using:
        ``__call__``, ``serial_request``
            .. automethod::
                cylc.flow.network.client.SuiteRuntimeClient.serial_request

        ``async_request``
            .. automethod::
                cylc.flow.network.client.SuiteRuntimeClient.async_request

    """

    DEFAULT_TIMEOUT = 5.  # 5 seconds

    def __init__(
            self,
            suite: str,
            owner: str = None,
            host: str = None,
            port: Union[int, str] = None,
            context: object = None,
            timeout: Union[float, str] = None,
            srv_public_key_loc: str = None
    ):
        super().__init__(zmq.REQ, context=context)
        self.suite = suite
        if port:
            port = int(port)
        if not (host and port):
            host, port, _ = get_location(suite, owner, host)
        if timeout is None:
            timeout = self.DEFAULT_TIMEOUT
        else:
            timeout = float(timeout)
        self.timeout = timeout * 1000
        self.timeout_handler = partial(
            self._timeout_handler, suite, host, port)
        self.poller = None
        # Connect the ZMQ socket on instantiation
        self.start(host, port, srv_public_key_loc)
        # gather header info post start
        self.header = self.get_header()

    def _socket_options(self):
        """Set socket options after socket instantiation before connect.

        Overwrites Base method.

        """
        # if there is no server don't keep the client hanging around
        self.socket.setsockopt(zmq.LINGER, int(self.DEFAULT_TIMEOUT))

        # create a poller to handle timeouts
        self.poller = zmq.Poller()
        self.poller.register(self.socket, zmq.POLLIN)

    async def async_request(self, command, args=None, timeout=None):
        """Send an asynchronous request using asyncio.

        Has the same arguments and return values as ``serial_request``.

        """
        timeout = (float(timeout) * 1000 if timeout else None) or self.timeout
        if not args:
            args = {}

        # Note: we are using CurveZMQ to secure the messages (see
        # self.curve_auth, self.socket.curve_...key etc.). We have set up
        # public-key cryptography on the ZMQ messaging and sockets, so
        # there is no need to encrypt messages ourselves before sending.

        # send message
        msg = {'command': command, 'args': args}
        msg.update(self.header)
        LOG.debug('zmq:send %s', msg)
        message = encode_(msg)
        self.socket.send_string(message)

        # receive response
        if self.poller.poll(timeout):
            res = await self.socket.recv()
        else:
            if callable(self.timeout_handler):
                self.timeout_handler()
            raise ClientTimeout(
                'Timeout waiting for server response.'
                ' This could be due to network or server issues.'
                ' Check the suite log.'
            )

        if msg['command'] in PB_METHOD_MAP:
            response = {'data': res}
        else:
            response = decode_(res.decode())
        LOG.debug('zmq:recv %s', response)

        try:
            return response['data']
        except KeyError:
            error = response['error']
            raise ClientError(error['message'], error.get('traceback'))

    def serial_request(self, command, args=None, timeout=None):
        """Send a request.

        For convenience use ``__call__`` to call this method.

        Args:
            command (str): The name of the endpoint to call.
            args (dict): Arguments to pass to the endpoint function.
            timeout (float): Override the default timeout (seconds).

        Raises:
            ClientTimeout: If a response takes longer than timeout to arrive.
            ClientError: Coverall for all other issues including failed auth.

        Returns:
            object: The data exactly as returned from the endpoint function,
                nothing more, nothing less.

        """
        task = self.loop.create_task(
            self.async_request(command, args, timeout))
        self.loop.run_until_complete(task)
        return task.result()

    @staticmethod
    def get_header() -> dict:
        """Return "header" data to attach to each request for traceability.

        Returns:
            dict: dictionary with the header information, such as
                program and hostname.
        """
        cmd = sys.argv[0]

        cylc_executable_location = which("cylc")
        if cylc_executable_location:
            cylc_bin_dir = os.path.abspath(
                os.path.join(cylc_executable_location, os.pardir)
            )
            if not cylc_bin_dir.endswith("/"):
                cylc_bin_dir = f"{cylc_bin_dir}/"

            if cmd.startswith(cylc_bin_dir):
                cmd = cmd.replace(cylc_bin_dir, '')

        return {
            'meta': {
                'prog': cmd,
                'host': socket.gethostname()
            }
        }

    @staticmethod
    def _timeout_handler(suite: str, host: str, port: Union[int, str]):
        """Handle the eventuality of a communication timeout with the suite.

        Args:
            suite (str): suite name
            host (str): host name
            port (Union[int, str]): port number
        Raises:
            ClientError: if the suite has already stopped.
        """
        if suite is None:
            return
        # Cannot connect, perhaps suite is no longer running and is leaving
        # behind a contact file?
        try:
            detect_old_contact_file(suite, (host, port))
        except (AssertionError, SuiteServiceFileError):
            # * contact file not have matching (host, port) to suite proc
            # * old contact file exists and the suite process still alive
            return
        else:
            # the suite has stopped
            raise SuiteStopped(suite)

    __call__ = serial_request



from textwrap import dedent

import pytest

from cylc.flow.context_node import ContextNode


with ContextNode('a') as a:
    with ContextNode('b') as b:
        c = ContextNode('c')
    d = ContextNode('d')


def test_context_node_tree():
    """Ensure parents and children are correctly organised."""
    assert a.name == 'a'
    assert a._parent is None
    assert a._children == {'b': b, 'd': d}
    assert b.name == 'b'
    assert b._parent == a
    assert b._children == {'c': c}
    assert c.name == 'c'
    assert c._parent == b
    assert c._children is None


def test_data_state():
    """Ensure the in-class state remains intact."""
    # DATA should start clean
    assert ContextNode.DATA == {}

    # nodes add themselves to DATA then remove themselves from it
    with ContextNode('foo'):
        pass

    # so DATA should end clean
    assert ContextNode.DATA == {}

    # however erroneously creating leaf-nodes outside of a context
    # will make a mess
    a = ContextNode('a')
    assert ContextNode.DATA == {a: set()}

    # just to make sure this doesn't result in hard-to-debug errors
    # we make sure this doesn't prevent us creating new trees
    with ContextNode('foo') as foo:
        bar = ContextNode('bar')
    assert ContextNode.DATA == {a: set(), foo: {a}}
    assert list(foo.walk()) == [(0, foo), (1, bar)]

    # finally clean up for niceness
    del ContextNode.DATA[a]
    del ContextNode.DATA[foo]


def test_context_iter():
    """Test iterating over child nodes."""
    assert list(a) == [b, d]
    assert list(b) == [c]
    assert list(c) == []


def test_context_contains():
    """Test `node.__contains__`."""
    assert 'b' in a
    assert 'c' in b
    assert 'c' not in a


def test_context_getitem():
    """Test `node.__getitem__`."""
    assert a['b'] == b
    assert b['c'] == c
    with pytest.raises(KeyError):
        assert b['d']
    with pytest.raises(TypeError):
        assert c['b']


def test_context_str():
    """Test the string representation of a node."""
    assert str(a) == 'a'
    assert str(b) == 'b'
    assert str(c) == 'c'


def test_context_repr():
    """Test the Python representation of a node."""
    assert repr(a) == 'a'
    assert repr(b) == 'a/b'
    assert repr(c) == 'a/b/c'


def test_context_walk():
    """Test walking the tree."""
    assert list(a.walk()) == [
        (0, a),
        (1, b),
        (2, c),
        (1, d)
    ]
    assert list(b.walk()) == [
        (0, b),
        (1, c),
    ]
    assert list(c.walk()) == [
        (0, c)
    ]


def test_context_walk_depth():
    assert list(a.walk(depth=1)) == [
        (0, a),
        (1, b),
        (1, d)
    ]


def test_context_tree():
    """Test the string representation of the tree."""
    assert a.tree() == dedent('''
        a
            b
                c
            d
    ''').strip()
    assert b.tree() == dedent('''
        b
            c
    ''').strip()
    assert c.tree() == dedent('''
        c
    ''').strip()


def test_context_is_root():
    """Test root node detection."""
    assert a.is_root()
    assert not b.is_root()
    assert not c.is_root()


def test_context_is_leaf():
    """Test leaf node detection."""
    assert not a.is_leaf()
    assert not b.is_leaf()
    assert c.is_leaf()


def test_context_parents():
    """Test linearised ancestry."""
    assert list(a.parents()) == []
    assert list(b.parents()) == [a]
    assert list(c.parents()) == [b, a]


def test_re_enter():
    """It should be able to re __enter__ multiple times."""
    # create a node with one child
    with ContextNode('x') as x:
        ContextNode('y')
    assert 'y' in x

    # go back and add another child retrospectively
    with x:
        ContextNode('z')
    assert 'y' in x
    assert 'z' in x



# THIS FILE IS PART OF THE CYLC SUITE ENGINE.
# Copyright (C) NIWA & British Crown (Met Office) & Contributors.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import pytest

from cylc.flow.batch_sys_handlers.moab import BATCH_SYS_HANDLER


@pytest.mark.parametrize(
    'job_conf,lines',
    [
        (  # basic
            {
                'batch_system_conf': {},
                'directives': {},
                'execution_time_limit': 180,
                'job_file_path': '$HOME/cylc-run/chop/log/job/1/axe/01/job',
                'suite_name': 'chop',
                'task_id': 'axe.1',
            },
            [
                '#PBS -N axe.1.chop',
                '#PBS -o cylc-run/chop/log/job/1/axe/01/job.out',
                '#PBS -e cylc-run/chop/log/job/1/axe/01/job.err',
                '#PBS -l walltime=180',
            ],
        ),
        (  # some useful directives
            {
                'batch_system_conf': {},
                'directives': {
                    '-q': 'forever',
                    '-V': '',
                    '-l mem': '256gb',
                },
                'execution_time_limit': 180,
                'job_file_path': '$HOME/cylc-run/chop/log/job/1/axe/01/job',
                'suite_name': 'chop',
                'task_id': 'axe.1',
            },
            [
                '#PBS -N axe.1.chop',
                '#PBS -o cylc-run/chop/log/job/1/axe/01/job.out',
                '#PBS -e cylc-run/chop/log/job/1/axe/01/job.err',
                '#PBS -l walltime=180',
                '#PBS -q forever',
                '#PBS -V',
                '#PBS -l mem=256gb',
            ],
        ),
    ],
)
def test_format_directives(job_conf: dict, lines: list):
    assert BATCH_SYS_HANDLER.format_directives(job_conf) == lines



# THIS FILE IS PART OF THE CYLC SUITE ENGINE.
# Copyright (C) NIWA & British Crown (Met Office) & Contributors.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""Run command on a remote, (i.e. a remote [user@]host)."""

import os
from shlex import quote
from posix import WIFSIGNALED
import shlex
import signal
# CODACY ISSUE:
#   Consider possible security implications associated with Popen module.
# REASON IGNORED:
#   Subprocess is needed, but we use it with security in mind.
from subprocess import Popen, PIPE, DEVNULL
import sys
from time import sleep

from cylc.flow import LOG
from cylc.flow.cfgspec.glbl_cfg import glbl_cfg
import cylc.flow.flags
from cylc.flow.hostuserutil import is_remote
from cylc.flow import __version__ as CYLC_VERSION


def get_proc_ancestors():
    """Return list of parent PIDs back to init."""
    pid = os.getpid()
    ancestors = []
    while True:
        p = Popen(["ps", "-p", str(pid), "-oppid="], stdout=PIPE, stderr=PIPE)
        ppid = p.communicate()[0].decode().strip()
        if not ppid:
            return ancestors
        ancestors.append(ppid)
        pid = ppid


def watch_and_kill(proc):
    """Kill proc if my PPID (etc.) changed - e.g. ssh connection dropped."""
    gpa = get_proc_ancestors()
    while True:
        sleep(0.5)
        if proc.poll() is not None:
            break
        if get_proc_ancestors() != gpa:
            sleep(1)
            os.kill(proc.pid, signal.SIGTERM)
            break


def run_cmd(
        command,
        stdin=None,
        stdin_str=None,
        capture_process=False,
        capture_status=False,
        manage=False
):
    """Run a given cylc command on another account and/or host.

    Arguments:
        command (list):
            command inclusive of all opts and args required to run via ssh.
        stdin (file):
            If specified, it should be a readable file object.
            If None, DEVNULL is set if output is to be captured.
        stdin_str (str):
            A string to be passed to stdin.
            Implies `stdin=PIPE`.
        capture_process (boolean):
            If True, set stdout=PIPE and return the Popen object.
        capture_status (boolean):
            If True, and the remote command is unsuccessful, return the
            associated exit code instead of exiting with an error.
        manage (boolean):
            If True, watch ancestor processes and kill command if they change
            (e.g. kill tail-follow commands when parent ssh connection dies).

    Return:
        * If capture_process=True, the Popen object if created successfully.
        * Else True if the remote command is executed successfully, or
          if unsuccessful and capture_status=True the remote command exit code.
        * Otherwise exit with an error message.
    """
    # CODACY ISSUE:
    #   subprocess call - check for execution of untrusted input.
    # REASON IGNORED:
    #   The command is read from the site/user global config file, but we check
    #   above that it ends in 'cylc', and in any case the user could execute
    #   any such command directly via ssh.
    stdout = None
    stderr = None
    if capture_process:
        stdout = PIPE
        stderr = PIPE
        if stdin is None:
            stdin = DEVNULL
    if stdin_str:
        read, write = os.pipe()
        os.write(write, stdin_str.encode())
        os.close(write)
        stdin = read

    try:
        proc = Popen(command, stdin=stdin, stdout=stdout, stderr=stderr)
    except OSError as exc:
        sys.exit(r'ERROR: %s: %s' % (
            exc, ' '.join(quote(item) for item in command)))

    if capture_process:
        return proc
    else:
        if manage:
            watch_and_kill(proc)
        res = proc.wait()
        if WIFSIGNALED(res):
            sys.exit(r'ERROR: command terminated by signal %d: %s' % (
                res, ' '.join(quote(item) for item in command)))
        elif res and capture_status:
            return res
        elif res:
            sys.exit(r'ERROR: command returns %d: %s' % (
                res, ' '.join(quote(item) for item in command)))
        else:
            return True


def construct_ssh_cmd(raw_cmd, user=None, host=None, forward_x11=False,
                      stdin=False, ssh_login_shell=None, ssh_cylc=None,
                      set_UTC=False, allow_flag_opts=False):
    """Append a bare command with further options required to run via ssh.

    Arguments:
        raw_cmd (list): primitive command to run remotely.
        user (string): user ID for the remote login.
        host (string): remote host name. Use 'localhost' if not specified.
        forward_x11 (boolean):
            If True, use 'ssh -Y' to enable X11 forwarding, else just 'ssh'.
        stdin:
            If None, the `-n` option will be added to the SSH command line.
        ssh_login_shell (boolean):
            If True, launch remote command with `bash -l -c 'exec "$0" "$@"'`.
        ssh_cylc (string):
            Location of the remote cylc executable.
        set_UTC (boolean):
            If True, check UTC mode and specify if set to True (non-default).
        allow_flag_opts (boolean):
            If True, check CYLC_DEBUG and CYLC_VERBOSE and if non-default,
            specify debug and/or verbosity as options to the 'raw cmd'.

    Return:
        A list containing a chosen command including all arguments and options
        necessary to directly execute the bare command on a given host via ssh.
    """
    command = shlex.split(glbl_cfg().get_host_item('ssh command', host, user))

    if forward_x11:
        command.append('-Y')
    if stdin is None:
        command.append('-n')

    user_at_host = ''
    if user:
        user_at_host = user + '@'
    if host:
        user_at_host += host
    else:
        user_at_host += 'localhost'
    command.append(user_at_host)

    # Pass CYLC_VERSION and optionally, CYLC_CONF_PATH & CYLC_UTC through.
    command += ['env', quote(r'CYLC_VERSION=%s' % CYLC_VERSION)]
    try:
        command.append(
            quote(r'CYLC_CONF_PATH=%s' % os.environ['CYLC_CONF_PATH']))
    except KeyError:
        pass
    if set_UTC and os.getenv('CYLC_UTC') in ["True", "true"]:
        command.append(quote(r'CYLC_UTC=True'))
        command.append(quote(r'TZ=UTC'))

    # Use bash -l?
    if ssh_login_shell is None:
        ssh_login_shell = glbl_cfg().get_host_item(
            'use login shell', host, user)
    if ssh_login_shell:
        # A login shell will always source /etc/profile and the user's bash
        # profile file. To avoid having to quote the entire remote command
        # it is passed as arguments to the bash script.
        command += ['bash', '--login', '-c', quote(r'exec "$0" "$@"')]

    # 'cylc' on the remote host
    if ssh_cylc:
        command.append(ssh_cylc)
    else:
        ssh_cylc = glbl_cfg().get_host_item('cylc executable', host, user)
        if ssh_cylc.endswith('cylc'):
            command.append(ssh_cylc)
        else:
            # TODO - raise appropriate exception
            raise ValueError(
                r'ERROR: bad cylc executable in global config: %s' % ssh_cylc)

    # Insert core raw command after ssh, but before its own, command options.
    command += raw_cmd

    if allow_flag_opts:
        if (cylc.flow.flags.verbose or os.getenv('CYLC_VERBOSE') in
                ["True", "true"]):
            command.append(r'--verbose')
        if (cylc.flow.flags.debug or os.getenv('CYLC_DEBUG') in
                ["True", "true"]):
            command.append(r'--debug')

    return command


def remote_cylc_cmd(
        cmd,
        user=None,
        host=None,
        stdin=None,
        stdin_str=None,
        ssh_login_shell=None,
        ssh_cylc=None,
        capture_process=False,
        manage=False
):
    """Run a given cylc command on another account and/or host.

    Arguments:
    Args are directly inputted to one of two functions; see those docstrings:
            * See 'construct_ssh_cmd()' docstring:
                * cmd (--> raw_cmd);
                * user;
                * host;
                * stdin;
                * ssh_login_shell;
                * ssh_cylc.
            * See 'run_cmd()' docstring:
                * stdin [see also above]
                * capture (--> capture_process);
                * manage.

    Return:
        If capture_process=True, return the Popen object if created
        successfully. Otherwise, return the exit code of the remote command.
    """
    return run_cmd(
        construct_ssh_cmd(
            cmd,
            user=user,
            host=host,
            stdin=True if stdin_str else stdin,
            ssh_login_shell=ssh_login_shell,
            ssh_cylc=ssh_cylc
        ),
        stdin=stdin,
        stdin_str=stdin_str,
        capture_process=capture_process,
        capture_status=True,
        manage=manage
    )


def remrun(dry_run=False, forward_x11=False, abort_if=None,
           set_rel_local=False):
    """Short for RemoteRunner().execute(...)"""
    return RemoteRunner().execute(
        dry_run, forward_x11, abort_if, set_rel_local)


class RemoteRunner(object):
    """Run current command on a remote host.

    If user/owner or host differ from username and localhost, strip the
    remote options from the commandline and reinvoke the command on the
    remote host by non-interactive ssh, then exit; else do nothing.

    To ensure that users are aware of remote re-invocation info is always
    printed, but to stderr so as not to interfere with results.
    """
    OPT_ARG_OPTS = ('--user', '--host', '--ssh-cylc')

    def __init__(self, argv=None):
        self.user = None  # i.e. owner; name it user for consistency with CLI
        self.host = None
        self.ssh_login_shell = None
        self.ssh_cylc = None
        self.argv = argv or sys.argv

        cylc.flow.flags.verbose = '-v' in self.argv or '--verbose' in self.argv

        # Detect and replace host and user options
        argv = self.argv[1:]

        self.args = []
        while argv:
            arg = argv.pop(0)
            if arg.startswith(tuple([opt + '=' for opt in self.OPT_ARG_OPTS])):
                # e.g. if arg is '--host=HOST' here set self.host to HOST
                opt_with_dashes, opt_arg = arg.split('=', 1)
                setattr(self, opt_with_dashes.strip('--'), opt_arg)
            elif arg in self.OPT_ARG_OPTS:  # if opt arg provided after a space
                # e.g. if arg is '--host' set self.host to next element in argv
                setattr(self, arg.strip('--'), argv.pop(0))
            elif arg == '--login':
                self.ssh_login_shell = True
            elif arg == '--no-login':
                self.ssh_login_shell = False
            else:
                self.args.append(arg)

        if self.user is None and self.host is None:
            self.is_remote = False
        else:
            self.is_remote = is_remote(self.host, self.user)

    def execute(self, dry_run=False, forward_x11=False, abort_if=None,
                set_rel_local=False):
        """Execute command on remote host.

        Returns False if remote re-invocation is not needed, True if it is
        needed and executes successfully otherwise aborts.
        """
        if not self.is_remote:
            return False

        if abort_if is not None and abort_if in sys.argv:
            sys.exit(
                "ERROR: option '%s' not available for remote run" % abort_if)

        cmd = [os.path.basename(self.argv[0])[5:]]  # /path/to/cylc-foo => foo
        for arg in self.args:
            cmd.append(quote(arg))
            # above: args quoted to avoid interpretation by the shell,
            # e.g. for match patterns such as '.*' on the command line.

        if set_rel_local:
            # State as relative localhost to prevent recursive host selection.
            cmd.append("--host=localhost")
        command = construct_ssh_cmd(
            cmd, user=self.user, host=self.host, forward_x11=forward_x11,
            ssh_login_shell=self.ssh_login_shell, ssh_cylc=self.ssh_cylc,
            set_UTC=True, allow_flag_opts=False)

        if dry_run:
            return command
        else:
            return run_cmd(command)



# THIS FILE IS PART OF THE CYLC SUITE ENGINE.
# Copyright (C) NIWA & British Crown (Met Office) & Contributors.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""Encapsulates polling activity for CLI commands."""

import sys
from time import sleep


class Poller(object):
    """Encapsulates polling activity for cylc commands. Derived classes
    must override the check() method to test the polling condition."""

    @classmethod
    def add_to_cmd_options(cls, parser, d_interval=60, d_max_polls=10):
        """Add command line options for commands that can do polling"""
        parser.add_option(
            "--max-polls",
            help="Maximum number of polls (default " + str(d_max_polls) + ").",
            metavar="INT",
            action="store",
            dest="max_polls",
            default=d_max_polls)
        parser.add_option(
            "--interval",
            help=(
                "Polling interval in seconds (default " + str(d_interval) +
                ")."
            ),
            metavar="SECS",
            action="store",
            dest="interval",
            default=d_interval)

    def __init__(self, condition, interval, max_polls, args):

        self.condition = condition  # e.g. "suite stopped"

        # check max_polls is an int
        try:
            self.max_polls = int(max_polls)
        except ValueError:
            sys.exit("max_polls must be an int")

        # check interval is an int
        try:
            self.interval = int(interval)
        except ValueError:
            sys.exit("interval must be an integer")

        self.n_polls = 0
        self.args = args  # any extra parameters needed by check()

    def check(self):
        """Abstract method. Test polling condition."""
        raise NotImplementedError()

    def poll(self):
        """Poll for the condition embodied by self.check().
        Return True if condition met, or False if polling exhausted."""

        if self.max_polls == 0:
            # exit 1 as we can't know if the condition is satisfied
            sys.exit("WARNING: nothing to do (--max-polls=0)")
        elif self.max_polls == 1:
            sys.stdout.write("checking for '%s'" % self.condition)
        else:
            sys.stdout.write("polling for '%s'" % self.condition)

        while self.n_polls < self.max_polls:
            self.n_polls += 1
            if self.check():
                sys.stdout.write(": satisfied\n")
                return True
            if self.max_polls > 1:
                sys.stdout.write(".")
                sleep(self.interval)
        sys.stdout.write("\n")
        if self.max_polls > 1:
            sys.stderr.write(
                "ERROR: condition not satisfied after %d polls\n" %
                self.max_polls)
        else:
            sys.stderr.write("ERROR: condition not satisfied\n")
        return False



# THIS FILE IS PART OF THE CYLC SUITE ENGINE.
# Copyright (C) NIWA & British Crown (Met Office) & Contributors.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""Log the number and size of each type of object in the data store.

.. note::

   This plugin is for Cylc developers debugging the data store.

If ``matplotlib`` is installed this plugin will plot results as a PDF in
the run directory when the suite is shut down (cleanly).

"""
import json
from pathlib import Path
from time import time

from cylc.flow.main_loop import (startup, shutdown, periodic)


try:
    import matplotlib
    matplotlib.use('Agg')
    from matplotlib import pyplot as plt
    PLT = True
except ModuleNotFoundError:
    PLT = False

from pympler.asizeof import asized


@startup
async def init(scheduler, state):
    """Construct the initial state."""
    state['objects'] = {}
    state['size'] = {}
    state['times'] = []
    for key, _ in _iter_data_store(scheduler.data_store_mgr.data):
        state['objects'][key] = []
        state['size'][key] = []


@periodic
async def log_data_store(scheduler, state):
    """Count the number of objects and the data store size."""
    state['times'].append(time())
    for key, value in _iter_data_store(scheduler.data_store_mgr.data):
        state['objects'][key].append(
            len(value)
        )
        state['size'][key].append(
            asized(value).size
        )


@shutdown
async def report(scheduler, state):
    """Dump data to JSON, attempt to plot results."""
    _dump(state, scheduler.suite_run_dir)
    _plot(state, scheduler.suite_run_dir)


def _iter_data_store(data_store):
    for item in data_store.values():
        for key, value in item.items():
            if key != 'workflow':
                yield (key, value)
        # there should only be one workflow in the data store
        break


def _dump(state, path):
    data = {
        'times': state['times'],
        'objects': state['objects'],
        'size': state['size']
    }
    json.dump(
        data,
        Path(path, f'{__name__}.json').open('w+')
    )
    return True


def _plot(state, path):
    if (
            not PLT
            or len(state['times']) < 2
    ):
        return False

    times = [tick - state['times'][0] for tick in state['times']]
    _, ax1 = plt.subplots(figsize=(10, 7.5))

    ax1.set_xlabel('Time (s)')

    ax1.set_ylabel('Objects')
    for key, objects in state['objects'].items():
        ax1.plot(times, objects, label=key)

    ax2 = ax1.twinx()
    ax2.set_ylabel('Size (kb)')
    for key, sizes in state['size'].items():
        ax2.plot(times, [x / 1000 for x in sizes], linestyle=':')

    ax1.legend(loc=0)
    ax2.legend(
        (ax1.get_children()[0], ax2.get_children()[0]),
        ('objects', 'size'),
        loc=1
    )

    # start the x-axis at zero
    ax1.set_xlim(0, ax1.get_xlim()[1])

    plt.savefig(
        Path(path, f'{__name__}.pdf')
    )
    return True



# THIS FILE IS PART OF THE CYLC SUITE ENGINE.
# Copyright (C) NIWA & British Crown (Met Office) & Contributors.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""Run reference test."""

from difflib import unified_diff
import re


from cylc.flow import LOG
from cylc.flow.exceptions import SuiteEventError
from cylc.flow.pathutil import get_suite_test_log_name


def run_reftest(config, ctx):
    """Run reference test at shutdown."""
    reffilename = config.get_ref_log_name()
    curfilename = get_suite_test_log_name(ctx.suite)
    ref = _load_reflog(reffilename)
    cur = _load_reflog(curfilename)
    if ref == cur:
        LOG.info('SUITE REFERENCE TEST PASSED')
    else:
        exc = SuiteEventError(
            'SUITE REFERENCE TEST FAILED\n'
            'triggering is NOT consistent with the reference log:\n%s\n'
            % '\n'.join(unified_diff(ref, cur, 'reference', 'this run'))
        )
        LOG.exception(exc)
        raise exc


def _load_reflog(filename):
    """Reference test: get trigger info from reference log."""
    res = []
    re_trig = re.compile(r'(\[.+\]\s-triggered\soff\s\[.+\])$')
    for line in open(filename, 'r'):
        match = re_trig.search(line)
        if match:
            res.append(match.groups()[0])
    res.sort()
    return res



# THIS FILE IS PART OF THE CYLC SUITE ENGINE.
# Copyright (C) NIWA & British Crown (Met Office) & Contributors.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
from unittest.mock import Mock

import pytest

from cylc.flow.exceptions import CylcError
from cylc.flow.main_loop.health_check import (
    _check_suite_run_dir,
    _check_contact_file
)


def test_check_suite_run_dir():
    """Ensure a missing suite run dir raises an CylcError."""
    sched = Mock(suite_run_dir='/a/b/c/d/e')
    with pytest.raises(CylcError):
        _check_suite_run_dir(sched)


def test_check_contact_file_data(monkeypatch):
    """Ensure differing contact file data raises CylcError."""
    contact_data = {
        'a': 'beef',
        'b': 2
    }
    sched = Mock(
        suite='foo',
        contact_data=dict(contact_data)
    )
    monkeypatch.setattr(
        'cylc.flow.main_loop.health_check.suite_files.load_contact_file',
        lambda x: dict(contact_data)
    )
    # pass
    _check_contact_file(sched)

    # fail
    contact_data['a'] = 'wellington'
    with pytest.raises(CylcError):
        _check_contact_file(sched)


def test_check_contact_file_io(monkeypatch):
    """Ensure IOError retrieving the contact file raises CylcError."""
    sched = Mock(suite='foo')

    def whoopsie(*_):
        raise IOError('')

    monkeypatch.setattr(
        'cylc.flow.main_loop.health_check.suite_files.load_contact_file',
        whoopsie
    )
    with pytest.raises(CylcError):
        _check_contact_file(sched)



# THIS FILE IS PART OF THE CYLC SUITE ENGINE.
# Copyright (C) NIWA & British Crown (Met Office) & Contributors.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""Standard pytest fixtures for unit tests."""
import pytest
from shutil import rmtree

from cylc.flow.cfgspec.globalcfg import SPEC
from cylc.flow.cycling.loader import (
    ISO8601_CYCLING_TYPE,
    INTEGER_CYCLING_TYPE
)
from cylc.flow.parsec.config import ParsecConfig


@pytest.fixture
def cycling_mode(monkeypatch):
    """Set the Cylc cycling mode."""
    def _cycling_mode(integer=True):
        monkeypatch.setattr(
            'cylc.flow.cycling.loader.DefaultCycler.TYPE',
            (INTEGER_CYCLING_TYPE if integer else ISO8601_CYCLING_TYPE)
        )
    return _cycling_mode


@pytest.fixture
def mock_glbl_cfg(tmp_path, monkeypatch):
    """A Pytest fixture for fiddling globalrc values.

    * Hacks the specified `glbl_cfg` object.
    * Can be called multiple times within a test function.

    Args:
        pypath (str):
            The python-like path to the global configuation object you want
            to fiddle.
            E.G. if you want to hack the `glbl_cfg` in
            `cylc.flow.scheduler` you would provide
            `cylc.flow.scheduler.glbl_cfg`
        rc_string (str):
            The globlal configuration as a multi-line string.

    Example:
        Change the value of `UTC mode` in the global config as seen from
        `the scheduler` module.

        def test_something(mock_glbl_cfg):
            mock_glbl_cfg(
                'cylc.flow.scheduler.glbl_cfg',
                '''
                    [cylc]
                        UTC mode = True
                '''
            )

    """
    # TODO: modify Parsec so we can use StringIO rather than a temp file.
    def _mock(pypath, rc_string):
        nonlocal tmp_path, monkeypatch
        global_rc_path = tmp_path / 'flow.rc'
        global_rc_path.write_text(rc_string)
        glbl_cfg = ParsecConfig(SPEC)
        glbl_cfg.loadcfg(global_rc_path)

        def _inner(cached=False):
            nonlocal glbl_cfg
            return glbl_cfg

        monkeypatch.setattr(pypath, _inner)

    yield _mock
    rmtree(tmp_path)



#!/usr/bin/env python3

# THIS FILE IS PART OF THE CYLC SUITE ENGINE.
# Copyright (C) NIWA & British Crown (Met Office) & Contributors.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""cylc [discovery] check-versions [OPTIONS] ARGS

Check the version of cylc invoked on each of SUITE's task host accounts when
CYLC_VERSION is set to *the version running this command line tool*.
Different versions are reported but are not considered an error unless the
-e|--error option is specified, because different cylc versions from 6.0.0
onward should at least be backward compatible.

It is recommended that cylc versions be installed in parallel and access
configured via the cylc version wrapper as described in the cylc INSTALL
file and User Guide. This must be done on suite and task hosts. Users then get
the latest installed version by default, or (like tasks) a particular version
if $CYLC_VERSION is defined.

Use -v/--verbose to see the command invoked to determine the remote version
(all remote cylc command invocations will be of the same form, which may be
site dependent -- see cylc global config documentation."""

import os
import sys
from time import sleep

import cylc.flow.flags
from cylc.flow.option_parsers import CylcOptionParser as COP
from cylc.flow.cylc_subproc import procopen
from cylc.flow import __version__ as CYLC_VERSION
from cylc.flow.config import SuiteConfig
from cylc.flow.subprocpool import SubProcPool
from cylc.flow.suite_files import parse_suite_arg
from cylc.flow.task_remote_mgr import TaskRemoteMgr
from cylc.flow.templatevars import load_template_vars
from cylc.flow.terminal import cli_function


def get_option_parser():
    parser = COP(__doc__, prep=True, jset=True)

    parser.add_option(
        "-e", "--error", help="Exit with error status "
        "if " + CYLC_VERSION + " is not available on all remote accounts.",
        action="store_true", default=False, dest="error")

    return parser


@cli_function(get_option_parser, remove_opts=['--host', '--user'])
def main(_, options, *args):
    # suite name or file path
    suite, suiterc = parse_suite_arg(options, args[0])

    # extract task host accounts from the suite
    config = SuiteConfig(
        suite,
        suiterc,
        options,
        load_template_vars(options.templatevars, options.templatevars_file))
    account_set = set()
    for name in config.get_namespace_list('all tasks'):
        account_set.add((
            config.get_config(['runtime', name, 'remote', 'owner']),
            config.get_config(['runtime', name, 'remote', 'host'])))
    task_remote_mgr = TaskRemoteMgr(suite, SubProcPool())
    for _, host_str in account_set:
        task_remote_mgr.remote_host_select(host_str)
    accounts = []
    while account_set:
        for user, host_str in account_set.copy():
            res = task_remote_mgr.remote_host_select(host_str)
            if res:
                account_set.remove((user, host_str))
                accounts.append((user, res))
        if account_set:
            task_remote_mgr.proc_pool.process()
            sleep(1.0)

    # Interrogate the each remote account with CYLC_VERSION set to our version.
    # Post backward compatibility concerns to do this we can just run:
    #   cylc version --host=HOST --user=USER
    # but this command only exists for version > 6.3.0.
    # So for the moment generate an actual remote invocation command string for
    # "cylc --version".

    # (save verbose flag as gets reset in remrun)
    verbose = cylc.flow.flags.verbose

    warn = {}
    contacted = 0
    for user, host in sorted(accounts):
        argv = ["cylc", "version"]
        if user and host:
            argv += ["--user=%s" % user, "--host=%s" % host]
            user_at_host = "%s@%s" % (user, host)
        elif user:
            argv += ["--user=%s" % user]
            user_at_host = "%s@localhost" % user
        elif host:
            argv += ["--host=%s" % host]
            user_at_host = host
        if verbose:
            print("%s: %s" % (user_at_host, ' '.join(argv)))
        proc = procopen(argv, stdin=open(os.devnull), stdoutpipe=True,
                        stderrpipe=True)
        out, err = proc.communicate()
        out = out.decode()
        err = err.decode()
        if proc.wait() == 0:
            if verbose:
                print("   %s" % out)
            contacted += 1
            out = out.strip()
            if out != CYLC_VERSION:
                warn[user_at_host] = out
        else:
            print('ERROR ' + user_at_host + ':', file=sys.stderr)
            print(err, file=sys.stderr)

    # report results
    if not warn:
        if contacted:
            print("All", contacted, "accounts have cylc-" + CYLC_VERSION)
    else:
        print("WARNING: failed to invoke cylc-%s on %d accounts:" % (
            CYLC_VERSION, len(warn)))
        m = max(len(ac) for ac in warn)
        for ac, warning in warn.items():
            print(' ', ac.ljust(m), warning)
        if options.error:
            sys.exit(1)


if __name__ == "__main__":
    main()



# THIS FILE IS PART OF THE CYLC SUITE ENGINE.
# Copyright (C) NIWA & British Crown (Met Office) & Contributors.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import unittest

from cylc.flow.cycling import (
    SequenceBase, IntervalBase, PointBase, parse_exclusion)


class TestBaseClasses(unittest.TestCase):
    """Test the abstract base classes cannot be instantiated on their own
    """

    def test_simple_abstract_class_test(self):
        """Cannot instantiate abstract classes, they must be defined in
        the subclasses"""
        self.assertRaises(TypeError, SequenceBase, "sequence-string",
                          "context_string")
        self.assertRaises(TypeError, IntervalBase, "value")
        self.assertRaises(TypeError, PointBase, "value")


class TestParseExclusion(unittest.TestCase):
    """Test cases for the parser function"""

    def test_parse_exclusion_simple(self):
        """Tests the simple case of exclusion parsing"""
        expression = "PT1H!20000101T02Z"
        sequence, exclusion = parse_exclusion(expression)

        self.assertEqual(sequence, "PT1H")
        self.assertEqual(exclusion, ['20000101T02Z'])

    def test_parse_exclusions_list(self):
        """Tests the simple case of exclusion parsing"""
        expression = "PT1H!(T03, T06, T09)"
        sequence, exclusion = parse_exclusion(expression)

        self.assertEqual(sequence, "PT1H")
        self.assertEqual(exclusion, ['T03', 'T06', 'T09'])

    def test_parse_exclusions_list_spaces(self):
        """Tests the simple case of exclusion parsing"""
        expression = "PT1H!    (T03, T06,   T09)   "
        sequence, exclusion = parse_exclusion(expression)

        self.assertEqual(sequence, "PT1H")
        self.assertEqual(exclusion, ['T03', 'T06', 'T09'])

    def test_parse_bad_exclusion(self):
        """Tests incorrectly formatted exclusions"""
        expression1 = "T01/PT1H!(T06, T09), PT5M"
        expression2 = "T01/PT1H!T03, PT17H, (T06, T09), PT5M"
        expression3 = "T01/PT1H! PT8H, (T06, T09)"
        expression4 = "T01/PT1H! T03, T06, T09"

        self.assertRaises(Exception, parse_exclusion, expression1)
        self.assertRaises(Exception, parse_exclusion, expression2)
        self.assertRaises(Exception, parse_exclusion, expression3)
        self.assertRaises(Exception, parse_exclusion, expression4)


if __name__ == "__main__":
    unittest.main()
